import os
import logging
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from .utils import extract_data_from_xml_file,filter_link_site_values,transform_df

def transform_and_load_extracted_data_from_ri_api(dir_path_xml_files : str,
                                                  output_parquet_path: str,
                                                  replace_existing_files: bool = True,
                                                  remove_extracted_files_after_done: bool = False,
                                                  logger: logging.Logger | None = None) -> bool:
    process_status = True

    if not replace_existing_files and os.path.exists(output_parquet_path):
        if logger:
            logger.warning(f'Pulando transformação, pois replace_existing_files={str(replace_existing_files)} e já existe um arquivo "{output_parquet_path}".')
        return process_status
    else:
        if os.path.exists(output_parquet_path):
            os.remove(output_parquet_path)
        xml_file_paths = [os.path.join(dir_path_xml_files, f)
                            for f in os.listdir(dir_path_xml_files)
                            if f.endswith('.xml')]
        if not xml_file_paths:
            if logger:
                logger.warning(f"Nenhum arquivo .xml encontrado no diretório: {dir_path_xml_files}")
            return False
        else:
            if logger:
                logger.info(f"Encontrados {len(xml_file_paths)} arquivos XML para processar.")

            writer = None  # Será inicializado no primeiro chunk válido

            try:
                for xml_file_path in xml_file_paths:
                    if logger:
                        logger.info(f"Processando arquivo: {xml_file_path}...")

                    file_records = extract_data_from_xml_file(xml_file_path,logger=logger)

                    if not file_records:
                        if logger:
                            logger.warning(f"Nenhum registro extraído de {os.path.basename(xml_file_path)}.")
                    else:
                        df = pd.DataFrame(file_records)
                        df['link_site'] = filter_link_site_values(df['link_site'].to_list())
                        df['source_xml_file'] = os.path.basename(xml_file_path)
                        
                        df = transform_df(df,logger)

                        table = pa.Table.from_pandas(df)

                        if writer is None:                            
                            writer = pq.ParquetWriter(output_parquet_path, table.schema)
                            if logger:
                                logger.info('Criação do arquivo parquet de dados transformados iniciada e primeiros dados inseridos')

                        writer.write_table(table)

                        if logger:
                            logger.info(f"{len(df)} registros gravados no arquivo parquet com dados transformados a partir de {os.path.basename(xml_file_path)}.")

                if writer:
                    writer.close()
                    if logger:
                        logger.info(f'Arquivo Parquet transformado e armazenado com sucesso em: "{output_parquet_path}"')
                    process_status = True
                else:
                    if logger:
                        logger.warning('Nenhum dado foi extraído de nenhum XML. Nenhum arquivo Parquet foi gerado.')
                    process_status = False

            except Exception as e:
                if logger:
                    logger.error(f"Erro durante o processamento: {e}", exc_info=True)
                process_status = False

            finally:
                if remove_extracted_files_after_done:
                    for file_path in xml_file_paths:
                        try:
                            os.remove(file_path)
                            if logger:
                                logger.info(f"Arquivo removido: {file_path}")
                        except Exception as e:
                            if logger:
                                logger.error(f'Erro ao remover "{file_path}": {e}', exc_info=True)
                return process_status



import pandas as pd

from . . . .common.for_strings import format_text

# https://estrutura.ufsc.br/centros-de-ensino/
# https://prograd.ufsc.br/cursos-de-graduacao-da-ufsc/
# https://propg.ufsc.br/cap-2/capdss/guia-dos-programas-de-pos-graduacao-da-ufsc/

# ? Info faltando
# ! Info inserida "a mais"

DIC_CAMPUS_CURSOS_CENTROS_SIGLAS = {'FLN':{'Propriedade Intelectual e Transferência de Tecnologia para Inovação'.upper():'CSE',
                                            'Licenciatura Intercultural Indígena do Sul da Mata Atlântica'.upper():'CFH',
                                            'Engenharia de Transportes e Gestão Territorial'.upper():'CTC',
                                            'Inglês: Estudos Linguísticos e Literários'.upper():'CCE',
                                            'Multicêntrico em Ciências Fisiológicas'.upper():'CCB',
                                            'Biologia Celular e do Desenvolvimento'.upper():'CCB',
                                            'Interdisciplinar em Ciências Humanas'.upper():'CFH',
                                            'Saúde Mental e Atenção Psicossocial'.upper():'CCS',
                                            'Biologia de Fungos, Algas e Plantas'.upper():'CCB',
                                            'Engenharia e Gestão do Conhecimento'.upper():'', # ?
                                            'Engenharia de Controle e Automação'.upper():'CTC',
                                            'Engenharia de Automação e Sistemas'.upper():'CTC',
                                            'Ciência e Tecnologia de Alimentos'.upper():'CCA',
                                            'Ciência e Engenharia de Materiais'.upper():'CTC',
                                            'Educação Científica e Tecnológica'.upper():'CED',
                                            'Planejamento e Controle de Gestão'.upper():'CSE',
                                            'Engenharia Sanitária e Ambiental'.upper():'CTC',
                                            'Gestão do Cuidado em Enfermagem'.upper():'CCS',
                                            'Sociologia e Ciência Política'.upper():'CFH',
                                            'Perícias Criminais Ambientais'.upper():'CCB',
                                            'Métodos e Gestão em Avaliação'.upper():'CTC',
                                            'Letras-Línguas Estrangeiras'.upper():'CCE',
                                            'Recursos Genéticos Vegetais'.upper():'CCA',
                                            'Administração Universitária'.upper():'CSE',
                                            'Biotecnologia e Biociências'.upper():'CCB',
                                            'Matemática Pura e Aplicada'.upper():'CFM',
                                            'Engenharia de Aquicultura'.upper():'CCA',
                                            'Assistência Farmacêutica'.upper():'CCS', 
                                            'Arquitetura e Urbanismo'.upper():'CTC', 
                                            'Engenharia de Alimentos'.upper():'CTC',
                                            'Relações Internacionais'.upper():'CSE',
                                            'Engenharia de Materiais'.upper():'CTC',
                                            'Engenharia de Produção'.upper():'CTC',
                                            'Secretariado Executivo'.upper():'CCE',
                                            'Sistemas de Informação'.upper():'CTC',
                                            'Ciências da Computação'.upper():'CTC',
                                            'Ciências dos Alimentos'.upper():'CCA',
                                            'Engenharia Eletrônica'.upper():'CTC',
                                            'Ciência da Informação'.upper():'CED',
                                            'Ciência da Computação'.upper():'CTC',
                                            'Ciência da Informação'.upper():'CED',
                                            'Informática em Saúde'.upper():'CCS', # ?
                                            'Engenharia Ambiental'.upper():'CTC',
                                            'Sociologia Política'.upper():'CFH', # !
                                            'Antropologia Social'.upper():'CFH',
                                            'Ciências Biológicas'.upper():'CCB',
                                            'Ciências Econômicas'.upper():'CSE',
                                            'Engenharia Elétrica'.upper():'CTC',
                                            'Engenharia Mecânica'.upper():'CTC',
                                            'Estudos da Tradução'.upper():'CCE',
                                            'Engenharia Química'.upper():'CTC',
                                            'Ensino de Biologia'.upper():'CCB',
                                            'Ciências Contábeis'.upper():'CSE',
                                            'Desastres Naturais'.upper():'CFH',
                                            'Engenharia Química'.upper():'CTC',
                                            'Ensino de História'.upper():'CFH',
                                            'Educação do Campo'.upper():'CED',
                                            'Ciências Médicas'.upper():'CCS',   
                                            'Ensino de Física'.upper():'CFM',
                                            'Letras-Português'.upper():'CCE',
                                            'Engenharia Civil'.upper():'CTC',
                                            'Ciências Sociais'.upper():'CFH',
                                            'Agroecossistemas'.upper():'CCA',
                                            'Biblioteconomia'.upper():'CED',
                                            'Educação Física'.upper():'CDS',
                                            'Fonoaudiologia'.upper():'CCS',
                                            'Serviço Social'.upper():'CSE',
                                            'Saúde Pública'.upper():'CCS', # !
                                            'Contabilidade'.upper():'CSE',
                                            'Administração'.upper():'CSE',
                                            'Biotecnologia'.upper():'CCB', # !
                                            'Letras-LIBRAS'.upper():'CCE',
                                            'Neurociências'.upper():'CCB',
                                            'Artes Cênicas'.upper():'CCE',
                                            'Antropologia'.upper():'CFH', 
                                            'Arquivologia'.upper():'CED',    
                                            'Oceanografia'.upper():'CFM',
                                            'Farmacologia'.upper():'CCS',
                                            'Meteorologia'.upper():'CFM',
                                            'Aquicultura'.upper():'CCA',  
                                            'Odontologia'.upper():'CCS',
                                            'Biociências'.upper():'CCB', # !
                                            'Linguística'.upper():'CCE',
                                            'Literatura'.upper():'CCE',
                                            'Matemática'.upper():'CFM',
                                            'Museologia'.upper():'CFH',
                                            'Psicologia'.upper():'CFH',
                                            'Enfermagem'.upper():'CCS',
                                            'Jornalismo'.upper():'CCE',
                                            'Bioquímica'.upper():'CCB',
                                            'Sociologia'.upper():'CFH', # !
                                            'Agronomia'.upper():'CCA',
                                            'Filosofia'.upper():'CFH',
                                            'Geografia'.upper():'CFH',
                                            'Pedagogia'.upper():'CED',
                                            'Zootecnia'.upper():'CCA',
                                            'Animação'.upper():'CCE',
                                            'Ecologia'.upper():'CCB',
                                            'Economia'.upper():'CSE',
                                            'Educação'.upper():'CED',
                                            'História'.upper():'CFH',
                                            'Geologia'.upper():'CFH',
                                            'Biologia'.upper():'CCB', # !
                                            'Farmácia'.upper():'CCS',
                                            'Medicina'.upper():'CCS',
                                            'Nutrição'.upper():'CCS',
                                            'Direito'.upper():'CCJ',
                                            'Química'.upper():'CFM',
                                            'Cinema'.upper():'CCE',
                                            'Design'.upper():'CCE',
                                            'Física'.upper():'CFM',
                                            'Letras'.upper():'CCE'},
                            'CUR':{'Medicina Veterinária Convencional e Integrativa'.upper():'CCR',
                                   'Ecossistemas Agrícolas e Naturais'.upper():'CCR',
                                   'Engenharia Florestal'.upper():'CCR',
                                   'Medicina Veterinária'.upper():'CCR',
                                   'Agronomia'.upper():'CCR'},
                            'JOI':{'Engenharia de Transportes e Logística'.upper():'CTJ',
                                   'Engenharia Ferroviária e Metroviária'.upper():'CTJ',
                                   'Bacharelado em Ciência e Tecnologia'.upper():'CTJ',
                                   'Engenharia Civil de Infraestrutura'.upper():'CTJ',                                   
                                   'Engenharia de Sistemas Eletrônicos'.upper():'CTJ',
                                   'Engenharia e Ciências Mecânicas'.upper():'CTJ',
                                   'Engenharia de Infraestrutura'.upper():'CTJ',
                                   'Engenharia Aeroespacial'.upper():'CTJ',
                                   'Engenharia Automotiva'.upper():'CTJ',
                                   'Engenharia Mecatrônica'.upper():'CTJ',
                                   'Engenharia Naval'.upper():'CTJ'},
                            'ARA':{'Tecnologias da Informação e Comunicação'.upper():'CTS',
                                   'Energia e Sustentabilidade'.upper():'CTS',
                                   'Ciências da Reabilitação'.upper():'CTS',
                                   'Engenharia de Computação'.upper():'CTS',
                                   'Engenharia de Energia'.upper():'CTS',
                                   'Ensino de Física'.upper():'CTS',
                                   'Fisioterapia'.upper():'CTS',
                                   'Medicina'.upper():'CTS'},
                            'BNU':{'Nanociência, Processos e Materiais Avançados'.upper():'CTE',
                                   'Química – Licenciatura e Bacharelado'.upper():'CTE',
                                   'Química - Licenciatura e Bacharelado'.upper():'CTE',
                                   'Engenharia de Controle e Automação'.upper():'CTE',
                                   'Química Licenciatura e Bacharelado'.upper():'CTE',
                                   'Matemática  Licenciatura'.upper():'CTE',
                                   'Engenharia de Materiais'.upper():'CTE',
                                   'Engenharia Têxtil'.upper():'CTE',
                                   'Ensino de Física'.upper():'CTE'}
}

DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS = {'FLN':{'Centro de Ciências Agrárias':'CCA',
                                               'Centro de Ciências Biológicas':'CCB',
                                               'Centro de Comunicação e Expressão':'CCE',
                                               'Centro de Ciências da Saúde':'CCS',
                                               'Centro de Ciências Jurídicas':'CCJ',
                                               'Centro de Desportos':'CDS',
                                               'Centro de Ciências da Educação':'CED',
                                               'Centro de Filosofia e Ciências Humanas':'CFH',
                                               'Centro de Ciências Físicas e Matemáticas':'CFM',
                                               'Centro Socioeconômico':'CSE',
                                               'Centro Tecnológico':'CTC'},
                                        'JOI':{'Centro Tecnológico de Joinville':'CTJ'},
                                        'CUR':{'Centro de Ciências Rurais':'CCR'},
                                        'BNU':{'Centro Tecnológico, de Ciências exatas e Educação':'CTE'},
                                        'ARA':{'Centro de Ciências, Tecnologias e Saúde':'CTS'}}

def format_cursos(curso : str) -> str:
    return format_text(curso).replace('(acadêmico)','').replace('(profissional)','').strip().upper()

# # def get_centros(areas : list[str],campus : list[str]) -> list[str]:
#     centros = []
#     for area,campi in zip(areas,campus):
#         if campi == 'FLN':
#             if area in ['Sociais Aplicadas']:
#                 centros.append('CSE')
#             elif area in ['Humanas']:
#                 centros.append('CFH')
#             elif area in ['Agrárias']:
#                 centros.append('CCA')
#             elif area in ['Saúde','Ciências da Saúde']:
#                 centros.append('CCS')
#             elif area in ['Biológicas']:
#                 centros.append('CCB')
#             elif area in ['Exatas e da Terra','Engenharias']:
#                 centros.append('CTC')
#             elif area in ['Linguística, Letras e Artes']:
#                 centros.append('CCE')
#             elif area in ['Multidisciplinar']:
#                 centros.append('Multidisciplinar')
#             else:
#                 centros.append('')
        
#         elif campi == 'CUR':
#             # if area in ['Medicina Veterinária']:
#             #     centros.append('CCR')
#             # elif area in ['Multidisciplinar']:
#             #     centros.append('Multidisciplinar')
#             # else:
#             #     centros.append('')
#             centros.append('CCR') # Curitibanos só tem um centro
        
#         elif campi == 'BNU':
#             # if area in ['Exatas e da Terra','Engenharias']:
#             #     centros.append('CTE')
#             # elif area in ['Multidisciplinar']:
#             #     centros.append('Multidisciplinar')
#             # else:
#             #     centros.append('')
#             centros.append('CTE') # Blumenau só tem um centro

#         elif campi == 'ARA':
#             centros.append('CTS') # Araranguá só tem um centro

#         elif campi == 'JOI':
#             centros.append('CTJ') # Joinville só tem um centro
        
#         else:
#             centros.append('')
#     return centros

def get_centros(cursos : list[str],
                campus : list[str]) -> list[str]:
    centros = []
    for curso,campi in zip(cursos,campus):
        if curso.strip() and campi.strip():
            try:
                centro = DIC_CAMPUS_CURSOS_CENTROS_SIGLAS[campi][curso]                
            except Exception as e:
                centros.append('')
            else:
                centros.append(centro)
        else:
            centros.append('')
    return centros

def format_campus(campus : str) -> str:
    if 'Florianópolis' in campus:
        return 'FLN'
    elif 'Blumenau' in campus:
        return 'BNU'
    elif 'Curitibanos' in campus:
        return 'CUR'
    elif 'Joinville' in campus:
        return 'JOI'
    elif 'Araranguá' in campus:
        return 'ARA'
    return ''

def insert_centro(df : pd.DataFrame) -> pd.DataFrame:
    df['CENTRO'] = get_centros(cursos=df['CURSO'].to_list(),
                               campus=df['CAMPUS'].to_list())
    return df




# import os
# import logging
# import pytz

# from datetime import datetime, timedelta
# from . .config import (
#     RESULTS_MAIN_DIR,LOGS_MAIN_DIR,UI_MAIN_DIR,
#     RESULTS_DIR_ETL_EXTRACTION,LOGS_DIR_ETL_EXTRACTION,
#     RESULTS_DIR_ETL_TRANSFORM_AND_LOAD,LOGS_DIR_ETL_TRANSFORM_AND_LOAD,
#     RESULTS_DIR_ETL_REPORT,LOGS_DIR_ETL_REPORT
# )

# def ensure_dir(path : str|list[str]):
#     if isinstance(path,str):
#         if not os.path.exists(path):
#             os.makedirs(path,exist_ok=True)
#     elif isinstance(path,list):
#         for p in [p for p in path if isinstance(p,str)]:
#             os.makedirs(p,exist_ok=True)

# def setup_dirs():        
#     ensure_dir([RESULTS_MAIN_DIR,
#                 LOGS_MAIN_DIR,
#                 UI_MAIN_DIR,
#                 RESULTS_DIR_ETL_EXTRACTION,
#                 LOGS_DIR_ETL_EXTRACTION,
#                 RESULTS_DIR_ETL_TRANSFORM_AND_LOAD,
#                 LOGS_DIR_ETL_TRANSFORM_AND_LOAD,                
#                 RESULTS_DIR_ETL_REPORT,
#                 LOGS_DIR_ETL_REPORT])

# def get_current_datetime() -> str:
#     brasilia_tz = pytz.timezone('America/Sao_Paulo')
#     brasilia_time = datetime.now(brasilia_tz)
#     return brasilia_time.strftime('d_%d_%m_%Y_t_%H_%M_%S')

# def get_old_loggings_file_paths(dir_path: str, max_days_living: int) -> list[str]:
#     brasilia_tz = pytz.timezone("America/Sao_Paulo")
#     now = datetime.now(tz=brasilia_tz)

#     log_files = [f for f in os.listdir(dir_path) if f.endswith('.log')]
#     files_to_delete = []

#     for log_file in log_files:
#         file_path = os.path.join(dir_path, log_file)

#         try:
#             # Pegando a data da última modificação
#             mod_timestamp = os.path.getmtime(file_path)
#             mod_datetime = datetime.fromtimestamp(mod_timestamp, tz=brasilia_tz)

#             # Verificando se o arquivo ultrapassou o tempo de vida
#             if now - mod_datetime > timedelta(days=max_days_living):
#                 files_to_delete.append(file_path)

#         except Exception as e:
#             pass

#     return files_to_delete
 

# def get_logger_for_current_run(loggings_full_file_path : str,
#                                remove_old_loggings : bool = True,
#                                max_days_living : int = 30):

#     if not loggings_full_file_path.endswith('.log'):
#         loggings_full_file_path += '.log'
    
#     if os.path.exists(loggings_full_file_path):
#         os.remove(loggings_full_file_path)
#     else:        
#         if os.path.dirname(loggings_full_file_path).strip():
#             ensure_dir(os.path.dirname(loggings_full_file_path))
    
#     if not os.path.dirname(loggings_full_file_path).strip():
#         loggings_full_file_path = os.path.join(os.getcwd(),loggings_full_file_path)
    
#     logging_config = {
#         'version': 1,
#         'disable_existing_loggers': False,
#         'formatters': {
#             'default': {
#                 'format': '%(asctime)s - %(levelname)s - %(message)s',
#                 'datefmt': '%d-%m-%Y %H:%M:%S',
#             },
#         },
#         'handlers': {
#             'file': {
#                 'class': 'logging.FileHandler',
#                 'filename': loggings_full_file_path,
#                 'level': 'INFO',
#                 'formatter': 'default',
#                 'encoding': 'utf-8',
#             },
#         },
#         'root': {
#             'handlers': ['file'],
#             'level': 'INFO',
#         },
#     }

#     # Configura o logger global
#     logging.config.dictConfig(logging_config)

#     # Retorna o logger para o módulo atual
#     logger = logging.getLogger(__name__)

#     if remove_old_loggings:
#         logger.info('Procurando arquivos antigos de log no diretório atual de log')
#         log_file_paths_to_delete = get_old_loggings_file_paths(dir_path=os.path.dirname(loggings_full_file_path),
#                                                                 max_days_living=max_days_living)
#         logger.info(f'Quantidade de arquivos de log para deletar: {len(log_file_paths_to_delete)}')
#         for log_file_path_to_delete in log_file_paths_to_delete:
#             os.remove(log_file_path_to_delete)

#     return logger



import pandas as pd

from .utils import DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS,DIC_CAMPUS_CURSOS_CENTROS_SIGLAS

class CursosUFSC():
    def __init__(self,silence : bool = True):
        self.dic_campus_centros_completo_e_siglas = DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS
        self.dic_campus_cursos_centros_siglas = DIC_CAMPUS_CURSOS_CENTROS_SIGLAS
        self.silence = silence        

    def get_cursos(self,
                   sort_by_len : bool = True,
                   reverse : bool = True) -> list[str]:
        cursos = []
        for campus in self.dic_campus_cursos_centros_siglas.keys():
            cursos += list(self.dic_campus_cursos_centros_siglas[campus].keys())
        if sort_by_len:
            return sorted(cursos,key=len,reverse=reverse)
        return list(cursos)
    
    def get_campus(self,
                   sort_by_len : bool = True,
                   reverse : bool = True) -> list[str]:        
        campus = [campi for campi in self.dic_campus_centros_completo_e_siglas.keys()]
        if sort_by_len:
            return sorted(campus,key=len,reverse=reverse)
        return list(campus)
    
    def get_centros(self,
                    siglas : bool = True,
                   sort_by_len : bool = True,
                   reverse : bool = True) -> list[str]:
        centros = []
        if siglas:
            for campi in self.dic_campus_centros_completo_e_siglas.keys():
                centros += [self.dic_campus_centros_completo_e_siglas[campi][centro] for centro in self.dic_campus_centros_completo_e_siglas[campi].keys()]
        else:
            for campi in self.dic_campus_centros_completo_e_siglas.keys():
                centros += list(self.dic_campus_centros_completo_e_siglas[campi].keys())
        if sort_by_len:
            return sorted(centros,key=len,reverse=reverse)
        return list(centros)



import pandas as pd
from . .common.for_strings import format_text

def clean_empty_rows(df: pd.DataFrame, columns: list[str]) -> pd.DataFrame:
    """
    Remove linhas com valores ausentes (NaN ou string vazia) nas colunas especificadas.
    """
    # Garante que os índices não estão duplicados
    df = df.copy()

    for col in columns:
        if col in df.columns:
            df = df.loc[df[col].notna() &(df[col] != '')].copy()
    
    df.reset_index(drop=True, inplace=True)
    return df


def filter_types(df: pd.DataFrame, types: list[str], exclude_empty_values: bool = True) -> pd.DataFrame:
    """
    Filtra o DataFrame com base na coluna 'type', mantendo apenas os registros cujo valor está na lista 'types'.
    Se 'include_empty' for True, também inclui linhas onde 'type' é string vazia "".
    """
    df = df.copy()
    # if not exclude_empty_values: # Comentando porque já foi tratado anteriormente
    #     df['type'].replace({'':'NÃO IDENTIFICADO'},inplace=True)
    # df = clean_empty_rows(df=df, columns=['type'])  # Assumindo que limpa NaN substituindo por ""
    
    if not exclude_empty_values:
        mask = df['type'].isin(types) | (df['type'] == 'NÃO ESPECIFICADO')
    else:
        mask = df['type'].isin(types)
    return df[mask]

def filter_dates(df: pd.DataFrame, date_1: int, date_2: int,exclude_empty_values : bool = True) -> pd.DataFrame:
    df = df.copy()
    
    if not exclude_empty_values:
        df_empty_values = df[(df['year']=='') | (df['year'].isna()) | (df['year'].isnull())]
        df_empty_values.loc[:, 'year'] = 'NÃO IDENTIFICADO'
        df = df.drop(df_empty_values.index)
    # Tenta converter para inteiro e remove valores inválidos
    df['year'] = pd.to_numeric(df['year'], errors='coerce')    
    df = clean_empty_rows(df=df,columns=['year'])  # Remove linhas que não puderam ser convertidas
    df['year'] = df['year'].astype(int)

    # Aplica filtro numérico
    filtered_df = df[(df['year'] >= date_1) & (df['year'] <= date_2)]

    if exclude_empty_values:
        return filtered_df
    else:
        return pd.concat([filtered_df,df_empty_values])

def filter_title_by_words(df: pd.DataFrame,
                          words: list[str],
                          match_all: bool = False,
                          exclude_empty_values: bool = True) -> pd.DataFrame:
    import re
    
    df = df.copy()

    # Isola os títulos vazios se for manter
    if not exclude_empty_values:
        df_empty_values = df[(df['title'] == '') | (df['title'].isna()) | (df['title'].isnull())].copy()
        df_empty_values.loc[:, 'title'] = 'NÃO IDENTIFICADO'
        df = df.drop(df_empty_values.index) # Remove do df principal para não aplicar filtro neles
    else:
        # Remove linhas com título vazio
        df = clean_empty_rows(df=df, columns=['title'])

    # Normaliza e escapa as palavras para regex
    words_formatted = [re.escape(format_text(word, special_treatment=True).strip()) for word in words]
    pattern = '|'.join(words_formatted)

    # Define a máscara de filtro
    if match_all:
        mask = df['title'].apply(
            lambda t: all(
                re.search(w, format_text(t, special_treatment=True), flags=re.IGNORECASE)
                for w in words_formatted
            )
        )
    else:
        mask = df['title'].apply(
            lambda t: re.search(pattern, format_text(t, special_treatment=True), flags=re.IGNORECASE) is not None
        )

    filtered = df[mask]

    if exclude_empty_values:
        return filtered
    else:
        # Retorna filtrados + os títulos originalmente vazios (agora com 'NÃO IDENTIFICADO')
        return pd.concat([filtered, df_empty_values])


def filter_subjects(df: pd.DataFrame,
                    subjects: list[str],
                    match_all: bool = False,
                    exclude_empty_values : bool = True) -> pd.DataFrame:
    df = df.copy()

    if not exclude_empty_values:
        df_empty_values = df[(df['subjects']=='') | (df['subjects'].isna()) | (df['subjects'].isnull())]
        df_empty_values.loc[:, 'subjects'] = 'NÃO IDENTIFICADO'
        df = df.drop(df_empty_values.index)
    else:
        # Remove linhas sem subjects
        df = clean_empty_rows(df=df,columns=['subjects'])

    # Normaliza os assuntos de entrada
    subjects_formatted = [format_text(s, special_treatment=True).strip() for s in subjects]

    # Função para verificar se algum assunto está presente
    def subject_match(row_subjects):
        row_subjects_split = row_subjects.split(';')
        row_subjects_formatted = [format_text(s.strip(), special_treatment=True) for s in row_subjects_split if s.strip()]
        if match_all:
            return all(any(word in subject for subject in row_subjects_formatted) for word in subjects_formatted)
        else:
            return any(any(word in subject for subject in row_subjects_formatted) for word in subjects_formatted)

    mask = df['subjects'].apply(subject_match)
    if exclude_empty_values:
        return df[mask]
    else:
        return pd.concat([df[mask],df_empty_values])

# Pega buscando "Igor Caetano" para authors = "Igor C Souza"
# Match all = True tem q conter todos os autores
def filter_authors(df: pd.DataFrame,
                   author_names: list[str],
                   match_all: bool = False,
                   exclude_empty_values : bool = True) -> pd.DataFrame:
    df = df.copy()

    if not exclude_empty_values:
        df_empty_values = df[(df['authors']=='') | (df['authors'].isna()) | (df['authors'].isnull())]
        df_empty_values.loc[:, 'authors'] = 'NÃO IDENTIFICADO'
        df = df.drop(df_empty_values.index)
    else:
        df = clean_empty_rows(df=df,columns=['authors'])

    # Normaliza e quebra os nomes buscados
    def extract_name_parts(name: str):
        words = [format_text(w.strip(), special_treatment=True) for w in name.split() if w.strip()]
        if not words:
            return None
        first = words[0]
        last = words[-1]
        last_initial = last[0] if last else ""
        return {"first": first, "last": last, "last_initial": last_initial}

    names_parts = [extract_name_parts(name) for name in author_names if name.strip()]

    def author_match(authors_raw):
        authors_list = authors_raw.split(';')
        authors_processed = [
            [format_text(part.strip(), special_treatment=True) for part in author.replace(',', ' ').split()]
            for author in authors_list
        ]

        matches = []
        for name in names_parts:
            if name is None:
                continue
            first, last, last_initial = name["first"], name["last"], name["last_initial"]

            matched = False
            for author_parts in authors_processed:
                has_first = first in author_parts
                has_last = (
                    last in author_parts or
                    f"{last_initial}." in author_parts or
                    last_initial in author_parts
                )
                if has_first and has_last:
                    matched = True
                    break

            matches.append(matched)

        return all(matches) if match_all else any(matches)

    mask = df['authors'].apply(author_match)
    if exclude_empty_values:
        return df[mask]
    else:
        return pd.concat([df[mask],df_empty_values])

# Pega buscando "Igor Caetano" para advisors = "Igor C Souza"
# Match all = True tem q conter todos os autores
def filter_advisors(df: pd.DataFrame,
                    advisor_names: list[str],
                    match_all: bool = False,
                    exclude_empty_values : bool = True) -> pd.DataFrame:
    df = df.copy()

    if not exclude_empty_values:
        df_empty_values = df[(df['advisors']=='') | (df['advisors'].isna()) | (df['advisors'].isnull())]
        df_empty_values.loc[:, 'advisors'] = 'NÃO IDENTIFICADO'
        df = df.drop(df_empty_values.index)
    else:
        df = clean_empty_rows(df=df,columns=['advisors'])

    # Normaliza e quebra os nomes buscados
    def extract_name_parts(name: str):
        words = [format_text(w.strip(), special_treatment=True) for w in name.split() if w.strip()]
        if not words:
            return None
        first = words[0]
        last = words[-1]
        last_initial = last[0] if last else ""
        return {"first": first, "last": last, "last_initial": last_initial}

    names_parts = [extract_name_parts(name) for name in advisor_names if name.strip()]

    def advisor_match(advisors_raw):
        advisors_list = advisors_raw.split(';')
        advisors_processed = [
            [format_text(part.strip(), special_treatment=True) for part in advisor.replace(',', ' ').split()]
            for advisor in advisors_list
        ]

        matches = []
        for name in names_parts:
            if name is None:
                continue
            first, last, last_initial = name["first"], name["last"], name["last_initial"]

            matched = False
            for advisor_parts in advisors_processed:
                has_first = first in advisor_parts
                has_last = (
                    last in advisor_parts or
                    f"{last_initial}." in advisor_parts or
                    last_initial in advisor_parts
                )
                if has_first and has_last:
                    matched = True
                    break

            matches.append(matched)

        return all(matches) if match_all else any(matches)

    mask = df['advisors'].apply(advisor_match)
    if exclude_empty_values:
        return df[mask]
    else:
        return pd.concat([df[mask],df_empty_values])

def filter_gender(df: pd.DataFrame,
                  genders: list[str],
                  just_contain: bool = True,
                  exclude_empty_values : bool = False) -> pd.DataFrame:
    df = df.copy()
    if not exclude_empty_values:
        df_empty_values = df[(df['gender_name']=='') | (df['gender_name'].isna()) | (df['gender_name'].isnull())]
        df_empty_values.loc[:, 'gender_name'] = 'NÃO IDENTIFICADO'
        df = df.drop(df_empty_values.index)
    else:
        df = clean_empty_rows(df=df,columns=['gender_name'])

    if just_contain:
        def match_any(g):
            if exclude_empty_values:
                return any(gender in g.split(',') for gender in genders)
            else:
                return any(gender in g.split(',') for gender in genders+[''])
        mask = df['gender_name'].apply(match_any)
    else:
        mask = df['gender_name'].isin(genders)

    if exclude_empty_values:
        return df[mask]
    else:
        return pd.concat([df[mask],df_empty_values])

def filter_language(df: pd.DataFrame,
                    languages: list[str],
                    exclude_empty_values : bool = False) -> pd.DataFrame:
    df = df.copy()

    # if exclude_empty_values:
    #     df = clean_empty_rows(df, columns=['language'])
    if not exclude_empty_values:
        mask = df['language'].isin(languages) | (df['language'] == "") | (df['language'].isna()) | (df['language'].isnull())
    else:
        mask = df['language'].isin(languages)
    # df = df[mask]
    if exclude_empty_values:
        return df[mask]
    else:
        df = df[mask]
        df['language'] = df['language'].replace({'': 'NÃO IDENTIFICADO'})
        return df

def filter_course(df: pd.DataFrame,
                  courses: list[str],
                  exclude_empty_values : bool = False) -> pd.DataFrame:
    df = df.copy()
    # if exclude_empty_values:
    #     df = clean_empty_rows(df=df,columns=['course'])
    if not exclude_empty_values:
        mask = df['course'].isin(courses) | (df['course'] == "") | (df['course'].isna()) | (df['course'].isnull())
    else:
        mask = df['course'].isin(courses)
    if exclude_empty_values:
        return df[mask]
    else:
        df = df[mask]
        df['course'] = df['course'].replace({'': 'NÃO IDENTIFICADO'})
        return df

def filter_type_course(df: pd.DataFrame, type_courses: list[str], exclude_empty_values: bool = False) -> pd.DataFrame:
    df = df.copy()
    # if exclude_empty_values:
    #     df = clean_empty_rows(df=df,columns=['type_course'])
    if not exclude_empty_values:
        mask = df['type_course'].isin(type_courses) | (df['type_course'] == "") | (df['type_course'].isna()) | (df['type_course'].isnull())
    else:
        mask = df['type_course'].isin(type_courses)
    if exclude_empty_values:
        return df[mask]
    else:
        df = df[mask]
        df['type_course'] = df['type_course'].replace({'': 'NÃO IDENTIFICADO'})
        return df

def filter_centro(df: pd.DataFrame, centros: list[str], exclude_empty_values: bool = False) -> pd.DataFrame:
    df = df.copy()
    # if exclude_empty_values:
    #     df = clean_empty_rows(df=df,columns=['centro'])
    if not exclude_empty_values:
        mask = df['centro'].isin(centros) | (df['centro'] == "") | (df['centro'].isna()) | (df['centro'].isnull())
    else:
        mask = df['centro'].isin(centros)
    if exclude_empty_values:
        return df[mask]
    else:
        df = df[mask]
        df['centro'] = df['centro'].replace({'': 'NÃO IDENTIFICADO'})
        return df

def filter_campus(df: pd.DataFrame, campuses: list[str], exclude_empty_values: bool = False) -> pd.DataFrame:
    df = df.copy()
    # if exclude_empty_values:
    #     df = clean_empty_rows(df=df,columns=['campus'])
    if not exclude_empty_values:
        mask = df['campus'].isin(campuses) | (df['campus'] == "") | (df['campus'].isna()) | (df['campus'].isnull())
    else:
        mask = df['campus'].isin(campuses)
    if exclude_empty_values:
        return df[mask]
    else:
        df = df[mask]
        df['campus'] = df['campus'].replace({'': 'NÃO IDENTIFICADO'})
        return df



import os

############################################         DIRECTORIES         ############################################

#####################    GENERAL    #####################

ROOT_DIR = os.getcwd()
# ROOT_DIR = r'C:\Users\Igor\Documents\GitHub\py_ri_ufsc_constru\Testings_py_ri_ufsc'
MAIN_DIR = os.path.join(ROOT_DIR,'py_ri_ufsc')
RESULTS_MAIN_DIR = os.path.join(MAIN_DIR,'results')
LOGS_MAIN_DIR = os.path.join(MAIN_DIR,'logs')
UI_MAIN_DIR = os.path.join(MAIN_DIR,'ui_files')


#####################    ETL    #####################
RESULTS_DIR_ETL = os.path.join(RESULTS_MAIN_DIR,'etl')
LOGS_DIR_ETL = os.path.join(LOGS_MAIN_DIR,'etl')
COL_TO_NAME_CSV_FILE_PATH = os.path.join(RESULTS_DIR_ETL,'col_to_name.csv')
###### EXTRACTION ######
RESULTS_DIR_ETL_EXTRACTION = os.path.join(RESULTS_DIR_ETL,'extraction')
LOGS_DIR_ETL_EXTRACTION = os.path.join(LOGS_DIR_ETL,'extraction')
###### TRANSFORM AND LOAD ######
RESULTS_DIR_ETL_TRANSFORM_AND_LOAD = os.path.join(RESULTS_DIR_ETL,'transform_and_load')
LOGS_DIR_ETL_TRANSFORM_AND_LOAD = os.path.join(LOGS_DIR_ETL,'transform_and_load')
COMPLETED_DATA_PARQUET_FILE_PATH = os.path.join(RESULTS_DIR_ETL_TRANSFORM_AND_LOAD,'tabled_data_completed.parquet')

###### REPORT ######
RESULTS_DIR_ETL_REPORT = os.path.join(RESULTS_DIR_ETL,'report')
LOGS_DIR_ETL_REPORT = os.path.join(LOGS_DIR_ETL,'report')

######################## USER INTERFACE ########################

UI_DOWNLOADS = os.path.join(UI_MAIN_DIR,'downloads')

############################################         CONSTANTS         ############################################

#####################    ETL    #####################
###### EXTRACTION ######
BASE_URL = 'https://repositorio.ufsc.br'
BASE_API_SINTAX = '/oai/request?'
BASE_VERB = 'ListRecords'
BASE_METADATA_PREFIX = 'xoai'
# EXPECTED_NUMBER_OF_ITEMS = 157021
# MAX_ITEMS_RETURNED_FROM_API = 100


this_dir,this_file = os.path.split(__file__)
DATASET_PARQUET_FILE_PATH = os.path.join(this_dir,'src_files','dataset.parquet')



import pandas as pd

from .get_metadata.utils import (
    get_available_values_in_dataset,get_available_columns_in_dataset,get_raw_dataset,
    get_filtered_raw_dataset
)
from .get_metadata.main import get_filtered_dataset_for_main_graphs
from .get_metadata.tests import TestRIUFSC


class RIUFSC():
    def __init__(self,
                 silence : bool = True):
        self.silence = silence
        if not silence:
            print('Não recomendamos carregar o dataset inteiro, tente sempre selecionar apenas as colunas que deseja utilizar.')

    def help(self):
        print('\n\n\tFunções disponíveis:\n')
        print('-'*100)
        print('get_available_columns_in_ri_ufsc_dataset -> lista todas as colunas disponíveis no dataset (RAM friendly).')
        print('-'*100)
        print('get_available_values_in_ri_ufsc_dataset -> lista todos os valores disponíveis na coluna desejada no dataset (RAM friendly).')
        print('-'*100)
        print('get_raw_ri_ufsc_dataset -> entrega um objeto DataFrame do pandas (selecione as colunas pelo parâmetro "columns_to_use" \nse não quiser carregar o dataset inteiro).')
        print('-'*100)
        print('get_filtered_raw_dataset_based_on_link_site_column -> entrega um objeto DataFrame do pandas com filtro de link_site aplicado \npara obter apenas registros que estejam presentes na lista "filter_links" (selecione as colunas que desejar \nno df de saída para minimizar o uso de RAM).')
        print('-'*100)
        print('get_filtered_ri_ufsc_dataset -> entrega um objeto DataFrame do pandas filtrado de acordo com os parâmetros selecionados.')
        print('-'*100)
        print('get_testing_object -> entrega um objeto TestRIUFSC para realização de testes com todo kit de ferramentas utilizado \nna construção/organização do dataset. Esse objeto tem funções próprias, veja o \nnotebook de testes disponibilizado no repositório.')
        print('-'*100)
        print('\n\n')

    def get_available_columns_in_ri_ufsc_dataset(self) -> list[str]:
        return get_available_columns_in_dataset()

    def get_available_values_in_ri_ufsc_dataset(self,column_name : str) -> list[str]:
        return get_available_values_in_dataset(column_name=column_name,silence=self.silence)

    def get_raw_ri_ufsc_dataset(self,columns_to_use : list[str]) -> pd.DataFrame:
        return get_raw_dataset(columns_to_use=columns_to_use)

    def get_filtered_raw_dataset_based_on_link_site_column(self,
                                                           columns_to_use : list[str],
                                                           filter_links : list[str]) -> pd.DataFrame:
        return get_filtered_raw_dataset(columns_to_use=columns_to_use,filter_links=filter_links)

    def get_filtered_ri_ufsc_dataset(self,
                                     type_filter : dict = {"use": False,"types":None,"exclude_empty_values":False},
                                     date_filter: dict = {"use": False, "date_1": None, "date_2": None,"exclude_empty_values":False},
                                     title_filter: dict = {"use": False, "words": None, "match_all": False,"exclude_empty_values":False},
                                     subjects_filter: dict = {"use": False, "subjects": None, "match_all": False,"exclude_empty_values":False},
                                     authors_filter: dict = {"use": False, "author_names": None, "match_all": False,"exclude_empty_values":False},
                                     advisors_filter: dict = {"use": False, "advisor_names": None, "match_all": False,"exclude_empty_values":False},
                                     gender_filter: dict = {"use": False, "genders": None, "just_contain": True,"exclude_empty_values":False},
                                     language_filter: dict = {"use": False, "languages": None, "exclude_empty_values": False},
                                     course_filter: dict = {"use": False, "courses": None, "exclude_empty_values": False},
                                     type_course_filter: dict = {"use": False, "type_courses": None, "exclude_empty_values": False},
                                     centro_filter: dict = {"use": False, "centros": None, "exclude_empty_values": False},
                                     campus_filter: dict = {"use": False, "campuses": None, "exclude_empty_values": False},
                                     exported_columns : list[str]|None = None) -> pd.DataFrame:
        return get_filtered_dataset_for_main_graphs(type_filter=type_filter,
                                                    date_filter=date_filter,
                                                    title_filter=title_filter,
                                                    subjects_filter=subjects_filter,
                                                    authors_filter=authors_filter,
                                                    advisors_filter=advisors_filter,
                                                    gender_filter=gender_filter,
                                                    language_filter=language_filter,
                                                    course_filter=course_filter,
                                                    type_course_filter=type_course_filter,
                                                    centro_filter=centro_filter,
                                                    campus_filter=campus_filter,
                                                    exported_columns=exported_columns,
                                                    silence=self.silence)

    def get_testing_object(self,
                           df : pd.DataFrame|None = None) -> TestRIUFSC:
        if not self.silence:
            print('A partir de agora use o objeto retornado como uma classe nova, específica para realização de testes.')
        return TestRIUFSC(df=df)



import pyarrow.parquet as pq
import pyarrow.dataset as ds
import pandas as pd
import pooch
import os

from . .config import DATASET_PARQUET_FILE_PATH

def download_ri_ufsc_dataset_via_hugging_face() -> str:
    print('\n\nEstamos baixando o dataset para que ele possa ser utilizado neste ambiente :)')
    print('Essa etapa de download só acontecerá uma vez...\n\n')
    
    url = "https://huggingface.co/datasets/igorcaetanods/ri_ufsc_dataset_2024/resolve/main/dataset.parquet" # URL direta para o arquivo no Hugging Face Hub

    filename = "dataset.parquet" # Nome local do arquivo (que será salvo na pasta aqui do pacote)

    path = pooch.retrieve(
        url=url,
        known_hash=None,
        fname=filename,
        path=os.path.dirname(DATASET_PARQUET_FILE_PATH),
        progressbar=True,
    )
    return path

def get_available_values_in_dataset(column_name : str,silence : bool = True) -> list[str]:
    try:
        df = pd.read_parquet(DATASET_PARQUET_FILE_PATH,columns=[column_name])
        counts = df[column_name].value_counts(dropna=True)        
    except Exception as e:
        if not silence:
            print(f'Erro na função get_available_values_in_dataset() --> {e}')
        return []
    else:
        return [f"{str(item)} ({count})" for item, count in counts.items()]

def get_raw_dataset(columns_to_use : list[str]) -> pd.DataFrame:
    if columns_to_use:
        return pd.read_parquet(DATASET_PARQUET_FILE_PATH,columns=columns_to_use)
    else:
        return pd.read_parquet(DATASET_PARQUET_FILE_PATH)

def get_available_columns_in_dataset() -> list[str]:
    """
    Retorna a lista de colunas de um arquivo Parquet sem carregar os dados na memória.

    Returns:
        list[str]: Lista com os nomes das colunas
    """
    parquet_file = pq.ParquetFile(DATASET_PARQUET_FILE_PATH)
    return parquet_file.schema.names

def get_filtered_raw_dataset(columns_to_use: list[str],
                             filter_links: list[str]) -> pd.DataFrame:
    """
    Carrega eficientemente um subconjunto de colunas e linhas de um Parquet grande,
    filtrando pelas colunas desejadas e valores de 'link_site'.

    Args:
        columns_to_use (list[str]): Colunas desejadas para exportação.
        filter_links (list[str]): Lista de 'link_site' a manter.

    Returns:
        pd.DataFrame: DataFrame resultante com as colunas e linhas filtradas.
    """
    # Garante que link_site está nas colunas lidas, pois é usado como filtro
    columns = list(set(columns_to_use + ['link_site']))

    # Define o dataset Parquet (pode ser particionado ou único)
    dataset = ds.dataset(DATASET_PARQUET_FILE_PATH, format="parquet")

    # Cria expressão de filtro para a coluna link_site
    filter_expr = ds.field("link_site").isin(filter_links)

    # Scanner eficiente: carrega só as colunas e linhas desejadas
    table = dataset.to_table(columns=columns, filter=filter_expr)

    df = table.to_pandas()

    # Remove 'link_site' se não estiver entre as colunas solicitadas
    if 'link_site' not in columns_to_use:
        df.drop(columns=['link_site'], inplace=True)

    return df


# def generate_export_dataset_file(columns_to_use : list[str],output_type : str = 'PARQUET'):
    
#     # Criando um uuid (uuid.uuid4() aleatório e único) e tornado-o mais curto em base64 (de 36 para 22 caracteres)
#     process_id = base64.urlsafe_b64encode(uuid.uuid4().bytes).rstrip(b'=').decode('ascii')

#     file_path_to_save = os.path.join(UI_DOWNLOADS,process_id)

#     df = get_raw_dataset(columns_to_use=columns_to_use)    
#     if output_type == 'PARQUET':        
#         df.to_parquet(file_path_to_save+'.parquet')# Botar caminho pra baixar o parquet e dps disponibilizar download
#     elif output_type == 'JSON':
#         df.to_json(file_path_to_save+'.json')
#     elif output_type == 'XLSX':
#         df.to_excel(file_path_to_save+'.xlsx',index=False)
#     elif output_type == 'CSV':
#         df.to_csv(file_path_to_save+'.csv',index=False)
    
#     return file_path_to_save



import re
import unicodedata
import ftfy
from string import punctuation, printable

LISTA_PUNCTS = punctuation.replace('-', '').replace('.', '')
NORMAL_CHARACTERS = printable + 'áàâãéèêíìîóòôõúùûüç' + 'áàâãéèêíìîóòôõúùûüç'.upper() + '–' + 'ª' + '°' + 'º'

STANDARD_SPECIAL_CHARACTERS_STRING = punctuation

def normalize_text(text: str) -> str:
    return ''.join(c for c in (d for char in text for d in unicodedata.normalize('NFD', char) if unicodedata.category(d) != 'Mn'))

def remove_strange_characters_from_text(text: str) -> str:
    # weird blank space that sometimes appears
    text = text.replace('\xa0', ' ')
    characters_to_remove = set([c for c in text if c not in NORMAL_CHARACTERS])
    # print('Removing:',characters_to_remove)
    for strange_character in characters_to_remove:
        text = text.replace(strange_character, '')
    return text

def remove_extra_blank_spaces_from_text(text: str) -> str:
    return re.sub(r'[^\S\n]+', ' ', text)

def remove_special_characters_from_text(text: str,
                                        string_special_characters: str = STANDARD_SPECIAL_CHARACTERS_STRING,
                                        remove_extra_blank_spaces: bool = True,
                                        remove_hiphen_btwn_words: bool = False,
                                        tratamento_personalizado: bool = True) -> str:
    if not remove_hiphen_btwn_words:
        string_special_characters = string_special_characters.replace('-', '')
        text = text.replace(' -', ' ').replace('- ', ' ')
    if not tratamento_personalizado:
        text = text.translate(str.maketrans('', '', string_special_characters))
    else:
        if remove_hiphen_btwn_words:
            string_special_characters_ad_espaco = r'\/\\\-'
        else:
            string_special_characters_ad_espaco = r'\/\\'
        text = text.translate(str.maketrans(
            string_special_characters_ad_espaco, ' '*len(string_special_characters_ad_espaco)))
        text = text.translate(str.maketrans('', '', string_special_characters))
    if remove_extra_blank_spaces:
        text = re.sub(r'\s+', ' ', text)
    return text.strip()


def format_text(text: str,
                lower_case: bool = False,
                normalize: bool = False,
                remove_special_characters: bool = False,
                string_special_characters: str = STANDARD_SPECIAL_CHARACTERS_STRING,
                remove_extra_blank_spaces: bool = True,
                remove_strange_characters: bool = True,
                special_treatment: bool = False) -> str:
    if special_treatment:
        lower_case = True
        normalize = True
        remove_special_characters = True
        string_special_characters = string_special_characters.replace('_', '')
        remove_extra_blank_spaces = True
        remove_strange_characters = True

    if remove_strange_characters:
        text = ftfy.fix_text(text)
        text = remove_strange_characters_from_text(text)
    if remove_special_characters:
        if special_treatment:
            text = remove_special_characters_from_text(text,
                                                       string_special_characters=string_special_characters,
                                                       remove_hiphen_btwn_words=True)
        else:
            text = remove_special_characters_from_text(text,
                                                       string_special_characters=string_special_characters)
    if remove_extra_blank_spaces:
        text = remove_extra_blank_spaces_from_text(text)
    if lower_case:
        text = text.lower()
    if normalize:
        text = normalize_text(text)
    if special_treatment:
        text = text.replace(' ', '_')
    return text



from os.path import exists
from . .config import DATASET_PARQUET_FILE_PATH

if not exists(DATASET_PARQUET_FILE_PATH):
    from .utils import download_ri_ufsc_dataset_via_hugging_face
    download_ri_ufsc_dataset_via_hugging_face()



import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from collections import Counter

def normalize_years_total(df: pd.DataFrame, year_col: str, year_range: tuple[int, int]) -> pd.DataFrame:
    """Retorna contagem por ano, preenchendo com 0 onde não houver registros."""
    all_years = pd.DataFrame({year_col: list(range(year_range[0], year_range[1] + 1))})
    actual_counts = df[year_col].value_counts().sort_index().reset_index()
    actual_counts.columns = [year_col, 'count']
    return all_years.merge(actual_counts, on=year_col, how='left').fillna(0).astype({year_col: int, 'count': int})

def plot_line_by_year(df: pd.DataFrame, year_col='year', year_range=(2002, 2025)):
        counts_by_year_df = normalize_years_total(df, year_col=year_col, year_range=year_range)
        fig = px.line(
            counts_by_year_df,
            x='year',
            y='count',
            markers=True,
            title='Quantidade total de Registros por Ano',
            labels={'year': 'Ano', 'count': 'Número de Registros'},
            template='plotly_white'
        )
        fig.update_layout(xaxis=dict(tickmode='linear', tickangle=-45))
        fig.update_traces(marker=dict(size=8, color='#007bff'), line=dict(width=2.5, color='#007bff'))
        return fig


def normalize_years_by_group(df: pd.DataFrame,
                              year_col: str,
                              group_col: str,
                              count_col_name: str = "count",
                              year_range: tuple[int, int] = None) -> pd.DataFrame:
    """
    Garante que todos os anos dentro do intervalo desejado estejam presentes em cada grupo.

    Args:
        df (pd.DataFrame): DataFrame contendo as colunas de ano e grupo.
        year_col (str): Nome da coluna de ano.
        group_col (str): Nome da coluna de agrupamento (ex: gênero).
        count_col_name (str): Nome da coluna de contagem. Default: "count".
        year_range (tuple[int, int], optional): Intervalo desejado de anos (min, max). 
                                                Se None, será inferido dos dados.

    Returns:
        pd.DataFrame: DataFrame com todos os anos presentes para cada grupo e contagens preenchidas.
    """
    # Agrupa e conta os valores
    counts_df = df.groupby([year_col, group_col]).size().reset_index(name=count_col_name)

    # Determina intervalo de anos
    if year_range is None:
        year_min = df[year_col].min()
        year_max = df[year_col].max()
    else:
        year_min, year_max = year_range

    all_years = list(range(year_min, year_max + 1))
    all_groups = counts_df[group_col].unique()

    # Cria todas as combinações possíveis de ano e grupo
    full_index = pd.MultiIndex.from_product([all_years, all_groups], names=[year_col, group_col]).to_frame(index=False)

    # Junta com os dados reais e preenche com 0 onde faltarem
    normalized = pd.merge(full_index, counts_df, how='left', on=[year_col, group_col]).fillna(0)
    normalized[count_col_name] = normalized[count_col_name].astype(int)

    return normalized

def plot_line_by_year_and_gender(df: pd.DataFrame, year_range=(2002, 2025)):
    df_gender = df[['year', 'gender_name']].copy()
    df_gender['year'] = df_gender['year'].astype(int)
    df_gender['gender_name'] = df_gender['gender_name'].replace({
        'F': 'Feminino', 'M': 'Masculino',
        'F,M': 'Feminino,Masculino', 'M,F': 'Masculino,Feminino',
        '': 'Não identificado'
    })
    df_gender = df_gender.assign(gender=df_gender['gender_name'].str.split(','))
    df_gender = df_gender.explode('gender')
    gender_counts = normalize_years_by_group(df_gender, year_col='year', group_col='gender', year_range=year_range)

    color_map = {'Feminino': '#9d3f96', 'Masculino': '#3f7f96', 'Não identificado': "#00FF0D"}
    fig = px.line(
        gender_counts,
        x='year',
        y='count',
        color='gender',
        markers=True,
        title='Quantidade de Registros por Ano e Gênero de autores',
        labels={'year': 'Ano', 'count': 'Número de Registros', 'gender': 'Gênero'},
        template='plotly_white',
        color_discrete_map=color_map
    )
    fig.update_layout(legend_title_text='Gênero dos autores', xaxis=dict(tickmode='linear', tickangle=-45))
    fig.update_traces(marker=dict(size=8), line=dict(width=2.5))
    return fig


def plot_language_pie(df: pd.DataFrame):
    lang_counts = df['language'].replace('', 'Não identificado').value_counts()
    color_map = {
        'por': "#832929", 'eng': '#3f7f96', 'Não identificado': '#00FF0D',
        'fra': "#8F1F85", 'ita': "#25BAFF", 'spa': "#9B8922",
        'deu': "#AC7522", 'mul': "#0014C9", 'zxx': "#585858",
        'bzs': "#B4B4B4", 'por;eng': "#665015"
    }
    cores_modernas = lang_counts.index.map(color_map)
    fig = go.Figure(data=[go.Pie(
        labels=lang_counts.index,
        values=lang_counts.values,
        hole=.4,
        pull=[0.05 if i == 0 else 0 for i in range(len(lang_counts.index))],
        marker_colors=cores_modernas,
        insidetextorientation='radial'
    )])
    fig.update_traces(
        textposition='inside',
        textinfo='percent+label',
        hovertemplate="<b>%{label}</b><br>Registros: %{value}<br>Porcentagem: %{percent}<extra></extra>",
        textfont_size=22
    )
    fig.update_layout(
        title={
            'text': '<b>Distribuição por Idioma</b>',
            'y': 0.95, 'x': 0.5, 'xanchor': 'center', 'yanchor': 'top', 'font': {'size': 20}
        },
        legend=dict(orientation="h", yanchor="bottom", y=-0.2, xanchor="center", x=0.5),
        annotations=[dict(text='<b>Idiomas</b>', x=0.5, y=0.5, font_size=22, showarrow=False)],
        uniformtext_minsize=10, uniformtext_mode='hide'
    )
    return fig


def plot_gender_pie(df: pd.DataFrame):
    df_gender_pizza = df['gender_name'].fillna('').astype(str).str.replace(' ', '').replace('', 'NI')
    df_gender_pizza = df_gender_pizza.str.split(',').explode()
    df_gender_pizza = df_gender_pizza.replace({'F': 'Feminino', 'M': 'Masculino', 'NI': 'Não identificado'})
    gender_counts = df_gender_pizza.value_counts().reset_index()
    gender_counts.columns = ['gender_label', 'count']

    color_map = {
        'Feminino': '#9d3f96', 'Masculino': '#3f7f96', 'Não identificado': '#00FF0D'
    }
    cores_ordenadas = gender_counts['gender_label'].map(color_map)
    fig = go.Figure(data=[go.Pie(
        labels=gender_counts['gender_label'],
        values=gender_counts['count'],
        hole=.4,
        pull=[0.05 if i == 0 else 0 for i in range(len(gender_counts.index))],
        marker_colors=cores_ordenadas,
        insidetextorientation='radial'
    )])
    fig.update_traces(
        textposition='inside',
        textinfo='percent+label',
        hovertemplate="<b>%{label}</b><br>Registros: %{value}<br>Porcentagem: %{percent}<extra></extra>",
        textfont_size=22
    )
    fig.update_layout(
        title={
            'text': '<b>Distribuição por Gênero</b>',
            'y': 0.95, 'x': 0.5, 'xanchor': 'center', 'yanchor': 'top', 'font': {'size': 20}
        },
        legend=dict(orientation="h", yanchor="bottom", y=-0.2, xanchor="center", x=0.5),
        annotations=[dict(text='<b>Gênero</b>', x=0.5, y=0.5, font_size=22, showarrow=False)],
        uniformtext_minsize=10, uniformtext_mode='hide'
    )
    return fig



def plot_top_subjects(
    df: pd.DataFrame,
    col_subjects: str = "subjects",
    top_n: int = 30,
    show_just_words: bool = False,
    remove_course_words: bool = False,
    stopwords_pt: set = None,
    stopwords_ufsc: set = None,
    format_text_func = None
):
    """
    Gera gráfico dos top assuntos (ou palavras) de um DataFrame.

    Parâmetros:
    - df: DataFrame contendo os dados
    - col_subjects: coluna que contém os assuntos (separados por ;)
    - top_n: número de itens mais frequentes a mostrar
    - show_just_words: se True, divide expressões compostas em palavras individuais
    - remove_course_words: se True, remove palavras relacionadas à UFSC
    - stopwords_pt: conjunto de palavras irrelevantes (português)
    - stopwords_ufsc: conjunto de palavras irrelevantes da UFSC
    - format_text_func: função usada para normalizar/verificar palavras (ex: format_text)

    Retorna:
    - fig: gráfico plotly com os assuntos
    - df_top: DataFrame com os assuntos e suas frequências
    """
    stopwords_pt = stopwords_pt or set()
    stopwords_ufsc = stopwords_ufsc or set()
    format_text_func = format_text_func or (lambda x, special_treatment=False: x)

    # Extrai assuntos
    assuntos_series = (
        df[col_subjects]
        .dropna()
        .astype(str)
        .str.split(';')
        .explode()
        .str.strip()
    )
    assuntos_series = assuntos_series[assuntos_series != ""]

    if show_just_words:
        assuntos_series = (
            assuntos_series
            .str.lower()
            .str.split()
            .explode()
            .str.strip()
        )
        assuntos_series = assuntos_series[
            (assuntos_series != "") & (~assuntos_series.isin(stopwords_pt))
        ]
        if remove_course_words:
            assuntos_series = [
                w for w in assuntos_series
                if len(w) > 1 and format_text_func(w, special_treatment=True) not in stopwords_ufsc
            ]
    elif remove_course_words:
        assuntos_series = (
            assuntos_series
            .str.lower()
            .str.strip()
        )
        assuntos_series = assuntos_series[
            (assuntos_series != "") & (~assuntos_series.isin(stopwords_pt))
        ]
        assuntos_series = [
            w for w in assuntos_series
            if len(w) > 1 and format_text_func(w, special_treatment=True) not in stopwords_ufsc
        ]

    contagem = Counter([
        item.upper()[:-1] if item and item[-1] in [',', ';', '.'] else item.upper()
        for item in assuntos_series
    ])

    df_top = (
        pd.Series(contagem)
        .nlargest(top_n)
        .reset_index()
        .rename(columns={'index': 'assunto', 0: 'frequencia'})
    )

    fig = px.bar(
        df_top.sort_values("frequencia"),
        x="frequencia",
        y="assunto",
        orientation="h",
        labels={"frequencia": "Frequência nos registros", "assunto": "Assunto"},
        color="frequencia",
        color_continuous_scale="Blues",
        title=f"Top {top_n} Assuntos"
    )
    fig.update_layout(yaxis=dict(tickfont=dict(size=10)))

    return df_top, fig

def plot_top_courses_by_year_and_subject(
    df: pd.DataFrame,
    subjects: list[str],
    match_all: bool = False,
    filter_subjects_func = None
):
    """
    Gera gráfico dos Top 3 cursos por ano para os assuntos especificados.

    Parâmetros:
    - df: DataFrame contendo colunas 'subjects', 'year' e 'course'
    - subjects: lista de assuntos a serem filtrados
    - match_all: se True, todos os assuntos devem estar presentes
    - filter_subjects_func: função que aplica o filtro nos dados

    Retorna:
    - fig: gráfico plotly
    - df_resultado: DataFrame usado no gráfico
    """
    if filter_subjects_func:
        df = filter_subjects_func(df=df, subjects=subjects, match_all=match_all)

    df = df.copy()
    df = df[df['year'].notna() & (df['year'] != "")]
    df['year'] = pd.to_numeric(df['year'], errors='coerce')
    df = df.dropna(subset=['year'])
    df = df[df['course'].notna() & (df['course'] != "")]

    if df.empty:
        return None, pd.DataFrame()

    grouped = df.groupby(['year', 'course']).size().reset_index(name='frequencia')
    top_cursos_ano = (
        grouped
        .sort_values(['year', 'frequencia'], ascending=[True, False])
        .groupby('year')
        .head(3)
    )

    anos = top_cursos_ano['year'].unique()
    cursos = top_cursos_ano['course'].unique()
    grid_completo = pd.MultiIndex.from_product([anos, cursos], names=["year", "course"]).to_frame(index=False)
    top_cursos_ano = pd.merge(grid_completo, top_cursos_ano, on=["year", "course"], how="left").fillna(0)
    top_cursos_ano['frequencia'] = top_cursos_ano['frequencia'].astype(int)

    fig = px.bar(
        top_cursos_ano,
        x="year",
        y="frequencia",
        color="course",
        barmode="group",
        labels={"frequencia": "Frequência", "year": "Ano", "course": "Curso"},
        title=f"Top 3 Cursos por ano com assunto{'s' if len(subjects) > 1 else ''}: {', '.join(subjects)}"
    )

    fig.update_layout(xaxis=dict(type='category'))

    return top_cursos_ano, fig



from setuptools import setup


with open(r'README.md','r',encoding='utf-8') as f:
    descricao_longa = f.read()

with open("requirements.txt") as f:
    install_requires = f.read().splitlines()

setup(
    name='py_ri_ufsc',
    version='0.4',
    # packages=find_packages(),
    packages=[
        "py_ri_ufsc",
        "py_ri_ufsc.common",
        "py_ri_ufsc.etl",
        "py_ri_ufsc.etl.extraction",
        "py_ri_ufsc.etl.extraction.courses_info",
        "py_ri_ufsc.etl.transform_and_load",
        "py_ri_ufsc.get_metadata",
        "py_ri_ufsc.src_files",
        "py_ri_ufsc.ui_graphs"
    ],
    package_dir={"": "src"},
    package_data={"py_ri_ufsc.src_files": ["*.parquet"]},
    include_package_data = True,
    install_requires = install_requires,
    description='Program design to handle searches in metadata stored in UFSC Institucional Repository',
    long_description=descricao_longa,
    long_description_content_type="text/markdown",
    author='Igor Caetano de Souza',
    project_urls={
        "GitHub Repository":"https://github.com/IgorCaetano/py_ri_ufsc"
    },
)




import pandas as pd

from .utils import get_raw_dataset
from .filters import (
    filter_types,filter_dates,filter_title_by_words,filter_subjects,
    filter_authors,filter_advisors,filter_gender,filter_language,
    filter_course,filter_type_course,filter_centro,filter_campus
)

def get_filtered_dataset_for_main_graphs(type_filter : dict = {"use": False,"types":None,"exclude_empty_values":False},
                                         date_filter: dict = {"use": False, "date_1": None, "date_2": None,"exclude_empty_values":False},
                                         title_filter: dict = {"use": False, "words": None, "match_all": False,"exclude_empty_values":False},
                                         subjects_filter: dict = {"use": False, "subjects": None, "match_all": False,"exclude_empty_values":False},
                                         authors_filter: dict = {"use": False, "author_names": None, "match_all": False,"exclude_empty_values":False},
                                         advisors_filter: dict = {"use": False, "advisor_names": None, "match_all": False,"exclude_empty_values":False},
                                         gender_filter: dict = {"use": False, "genders": None, "just_contain": True,"exclude_empty_values":False},
                                         language_filter: dict = {"use": False, "languages": None, "exclude_empty_values": False},
                                         course_filter: dict = {"use": False, "courses": None, "exclude_empty_values": False},
                                         type_course_filter: dict = {"use": False, "type_courses": None, "exclude_empty_values": False},
                                         centro_filter: dict = {"use": False, "centros": None, "exclude_empty_values": False},
                                         campus_filter: dict = {"use": False, "campuses": None, "exclude_empty_values": False},
                                         exported_columns : list[str]|None = None,
                                         silence : bool = True) -> pd.DataFrame:
    
    columns_to_use = ['year','gender_name','language','link_site']
    if not silence:
        print(f'Colunas fixas para uso: {str(columns_to_use)}...')
    if type_filter.get('use',False):
        columns_to_use.append('type')
        if not silence:
            print('Adicionando coluna type...')
    # if date_filter.get('use',False):
    #     columns_to_use.append('year')
    if title_filter.get('use',False):
        columns_to_use.append('title')
        if not silence:
            print('Adicionando coluna title...')
    if subjects_filter.get('use',False):
        columns_to_use.append('subjects')
        if not silence:
            print('Adicionando coluna subjects...')
    if authors_filter.get('use',False):
        columns_to_use.append('authors')
        if not silence:
            print('Adicionando coluna authors...')
    if advisors_filter.get('use',False):
        columns_to_use.append('advisors')
        if not silence:
            print('Adicionando coluna advisors...')
    # if gender_filter.get('use',False):
    #     columns_to_use.append('gender_name')
    # if language_filter.get('use',False):
    #     columns_to_use.append('language')
    if course_filter.get('use',False):
        columns_to_use.append('course')
        if not silence:
            print('Adicionando coluna course...')
    if type_course_filter.get('use',False):
        columns_to_use.append('type_course')
        if not silence:
            print('Adicionando coluna type_course...')
    if centro_filter.get('use',False):
        columns_to_use.append('centro')
        if not silence:
            print('Adicionando coluna centro...')
    if campus_filter.get('use',False):
        columns_to_use.append('campus')
        if not silence:
            print('Adicionando coluna campus...')

    if exported_columns:
        if not silence:
            print('Obtendo dataframe via dataset com colunas para uso + colunas para exportação...')
        df = get_raw_dataset(columns_to_use=list(set(columns_to_use+exported_columns)))
    else:
        if not silence:
            print('Obtendo dataframe via dataset com colunas para uso...')
        df = get_raw_dataset(columns_to_use=columns_to_use)

    df_filtered = df.copy().reset_index().drop(columns=['index'])
    if not silence:
        print('Iniciando filtragem do dataframe...')
    # Datas
    if date_filter and date_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de datas...')
        df_filtered = filter_dates(
            df=df_filtered,
            date_1=date_filter["date_1"],
            date_2=date_filter["date_2"],
            exclude_empty_values=date_filter['exclude_empty_values']
        )

    # Tipos
    if type_filter and type_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de tipos...')
        df_filtered = filter_types(
            df=df_filtered,
            types=type_filter["types"],
            exclude_empty_values=type_filter["exclude_empty_values"]
        )

    # Language
    if language_filter and language_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de idiomas...')
        df_filtered = filter_language(
            df=df_filtered,
            languages=language_filter["languages"],
            exclude_empty_values=language_filter['exclude_empty_values']
        )

    # Centro
    if centro_filter and centro_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de centros...')
        df_filtered = filter_centro(
            df=df_filtered,
            centros=centro_filter["centros"],
            exclude_empty_values=centro_filter["exclude_empty_values"]
        )

    # Campus
    if campus_filter and campus_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de campus...')
        df_filtered = filter_campus(
            df=df_filtered,
            campuses=campus_filter["campuses"],
            exclude_empty_values=campus_filter["exclude_empty_values"]
        )

    # Course
    if course_filter and course_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de cursos...')
        df_filtered = filter_course(
            df=df_filtered,
            courses=course_filter["courses"],
            exclude_empty_values=course_filter["exclude_empty_values"]
        )

    # Type Course
    if type_course_filter and type_course_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de tipos de curso...')
        df_filtered = filter_type_course(
            df=df_filtered,
            type_courses=type_course_filter["type_courses"],
            exclude_empty_values=type_course_filter["exclude_empty_values"]
        )

    # Gender
    if gender_filter and gender_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de gênero dos autores...')
        df_filtered = filter_gender(
            df=df_filtered,
            genders=gender_filter["genders"],
            just_contain=gender_filter["just_contain"],
            exclude_empty_values=gender_filter['exclude_empty_values']
        )

    # Título
    if title_filter and title_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de títulos...')
        df_filtered = filter_title_by_words(
            df=df_filtered,
            words=title_filter["words"],
            match_all=title_filter["match_all"],
            exclude_empty_values=title_filter["exclude_empty_values"]
        )

    # Subjects
    if subjects_filter and subjects_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de assuntos...')
        df_filtered = filter_subjects(
            df=df_filtered,
            subjects=subjects_filter["subjects"],
            match_all=subjects_filter["match_all"],
            exclude_empty_values=subjects_filter['exclude_empty_values']
        )

    # Authors
    if authors_filter and authors_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de autores...')
        df_filtered = filter_authors(
            df=df_filtered,
            author_names=authors_filter["author_names"],
            match_all=authors_filter["match_all"],
            exclude_empty_values=authors_filter['exclude_empty_values']
        )

    # Advisors
    if advisors_filter and advisors_filter.get("use", False):
        if not silence:
            print('Iniciando filtro de orientadores...')
        df_filtered = filter_advisors(
            df=df_filtered,
            advisor_names=advisors_filter["advisor_names"],
            match_all=advisors_filter["match_all"],
            exclude_empty_values=advisors_filter["exclude_empty_values"]
        )

    return df_filtered



import pandas as pd
import os
import logging
import requests
from bs4 import BeautifulSoup
import csv
import re

from lxml import etree
from br_gender.base import br_gender_info
from py_ri_ufsc.config import COL_TO_NAME_CSV_FILE_PATH
from py_ri_ufsc.common.for_strings import format_text
from py_ri_ufsc.etl.extraction.courses_info import CursosUFSC,DIC_CAMPUS_CURSOS_CENTROS_SIGLAS,DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS

# Defina os namespaces globalmente para serem usados em todas as funções XPath
NAMESPACES = {
    'oai': 'http://www.openarchives.org/OAI/2.0/',
    'xoai': 'http://www.lyncode.com/xoai'
}

def format_list_to_string(data_list):
    """Converte uma lista de strings em uma única string separada por ';', ou retorna None."""
    if data_list and isinstance(data_list, list):
        cleaned_list = [str(item).strip() for item in data_list if str(item).strip()]
        if not cleaned_list:
            return None
        return ';'.join(cleaned_list)
    elif isinstance(data_list, str) and data_list.strip():
        return data_list.strip()
    return None

def get_date_field(element, date_type_name):
    """Extrai um tipo específico de data (issued, available, accessioned)."""
    for lang_code in ['pt_BR', 'none', 'en']:
        date_elements = element.xpath(
            f"./xoai:element[@name='dc']/xoai:element[@name='date']/xoai:element[@name='{date_type_name}']/xoai:element[@name='{lang_code}']/xoai:field[@name='value']/text()",
            namespaces=NAMESPACES
        )
        if date_elements:
            dates = [date.strip() for date in date_elements if date.strip()]
            return dates[0] if dates else None
    
    date_elements = element.xpath(
        f"./xoai:element[@name='dc']/xoai:element[@name='date']/xoai:element[@name='{date_type_name}']/xoai:element/xoai:field[@name='value']/text()",
        namespaces=NAMESPACES
    )
    if date_elements:
        dates = [date.strip() for date in date_elements if date.strip()]
        return dates[0] if dates else None
    return None

def get_specific_dc_field(element, dc_element_name, sub_elements=None):
    """
    Extrai um campo específico do DC, com capacidade de navegar por sub_elements,
    priorizando pt_BR, none, en, ou o primeiro valor encontrado.
    Retorna uma string unida por ';' se múltiplos valores forem encontrados, ou uma única string.
    """
    base_xpath = f"./xoai:element[@name='dc']/xoai:element[@name='{dc_element_name}']"
    if sub_elements:
        for sub in sub_elements:
            base_xpath += f"/xoai:element[@name='{sub}']"

    collected_values = []
    for lang_code in ['pt_BR', 'none', 'en']:
        field_elements = element.xpath(
            f"{base_xpath}/xoai:element[@name='{lang_code}']/xoai:field[@name='value']/text()",
            namespaces=NAMESPACES
        )
        if field_elements:
            collected_values.extend([val.strip() for val in field_elements if val.strip()])
            if collected_values: break

    if not collected_values:
        field_elements = element.xpath(
            f"{base_xpath}/xoai:element/xoai:field[@name='value']/text()",
            namespaces=NAMESPACES
        )
        if field_elements:
            collected_values.extend([val.strip() for val in field_elements if val.strip()])

    if not collected_values:
        field_elements = element.xpath(
            f"{base_xpath}/xoai:field[@name='value']/text()",
            namespaces=NAMESPACES
        )
        if field_elements:
            collected_values.extend([val.strip() for val in field_elements if val.strip()])
            
    return format_list_to_string(list(set(collected_values))) if collected_values else None

def get_main_description_field(metadata_element):
    """
    Extrai o texto de descrição principal, excluindo o conteúdo do abstract.
    Prioriza pt_BR, depois none, depois en, ou texto direto.
    """
    description_dc_element = metadata_element.xpath(
        "./xoai:element[@name='dc']/xoai:element[@name='description']",
        namespaces=NAMESPACES
    )
    if not description_dc_element:
        return None
    
    description_node = description_dc_element[0]
    description_texts = []

    for lang_code in ['pt_BR', 'none', 'en']:
        lang_elements = description_node.xpath(
            f"./xoai:element[@name='{lang_code}' and not(local-name()='element' and @name='abstract')]/xoai:field[@name='value']/text()",
            namespaces=NAMESPACES
        )
        if lang_elements:
            description_texts.extend([text.strip() for text in lang_elements if text.strip()])

    direct_field_texts = description_node.xpath("./xoai:field[@name='value']/text()", namespaces=NAMESPACES)
    description_texts.extend([text.strip() for text in direct_field_texts if text.strip()])
    
    return format_list_to_string(list(set(description_texts))) if description_texts else None

def get_contributor_field_values(element, contributor_type_name):
    """Extrai valores de campos de contribuidores (author, advisor, advisor-co)."""
    base_contributor_xpath = f"./xoai:element[@name='dc']/xoai:element[@name='contributor']/xoai:element[@name='{contributor_type_name}']"
    specific_elements = element.xpath(base_contributor_xpath, namespaces=NAMESPACES)
    values_list = []
    for el in specific_elements:
        found_in_lang_specific = False
        for lang_code in ['pt_BR', 'none', 'en']:
            lang_specific_values = el.xpath(
                f"./xoai:element[@name='{lang_code}']/xoai:field[@name='value']/text()",
                namespaces=NAMESPACES
            )
            if lang_specific_values:
                values_list.extend([v.strip() for v in lang_specific_values if v.strip()])
                found_in_lang_specific = True
        if not found_in_lang_specific:
            direct_values = el.xpath("./xoai:field[@name='value']/text()", namespaces=NAMESPACES)
            if direct_values:
                values_list.extend([v.strip() for v in direct_values if v.strip()])
            else:
                # Fallback para encontrar qualquer campo de valor, caso a estrutura seja inesperada
                generic_values = el.xpath(".//xoai:field[@name='value']/text()", namespaces=NAMESPACES)
                if generic_values:
                    values_list.extend([v.strip() for v in generic_values if v.strip()])
    return list(set(values_list)) if values_list else []

def filter_link_site_values(list_of_values : list[str]) -> list[str]:
    filtered_values = []
    for item in list_of_values:
        if item:
            links = item.split(';')
            # Procura o link com domínio específico
            preferred = next((link for link in links if 'repositorio.ufsc.br' in link), None)
            # Se encontrou, usa o preferido; se não, usa o último da lista
            filtered_values.append(preferred if preferred else links[-1])
        else:
            filtered_values.append('')
    return filtered_values

def get_text_xpath(element, xpath_expr):
    """
    Retorna o primeiro valor de texto que encontrar usando a expressão XPath,
    ou None se não encontrar nada.
    """
    if element is None:
        return None
    results = element.xpath(xpath_expr, namespaces=NAMESPACES)
    if results:
        return results[0].strip() if results[0] else None
    return None

def extract_identifier(header, _): 
    return get_text_xpath(header, './oai:identifier/text()')

def extract_datestamp(header, _): 
    return get_text_xpath(header, './oai:datestamp/text()')

def extract_set_spec(header, _):
    specs = header.xpath('./oai:setSpec/text()', namespaces=NAMESPACES)
    for spec in specs:
        if spec.strip().startswith("col_"):
            return spec.strip()
    return None

def extract_authors(_, metadata): 
    return format_list_to_string(get_contributor_field_values(metadata, "author"))

def extract_advisors(_, metadata): 
    return format_list_to_string(get_contributor_field_values(metadata, "advisor"))

def extract_co_advisors(_, metadata): 
    return format_list_to_string(get_contributor_field_values(metadata, "advisor-co"))

def extract_link_doc(_, metadata):
    if metadata is None:
        return None
    links = metadata.xpath(
        ".//xoai:element[@name='bundles']/xoai:element[@name='bundle']/xoai:element[@name='bitstreams']"
        "/xoai:element[@name='bitstream' and xoai:field[@name='format']='application/pdf']"
        "/xoai:field[@name='url']/text()",
        namespaces=NAMESPACES
    )
    return format_list_to_string(list(set(link.strip() for link in links if link.strip())))

def extract_language(_, metadata):
    return get_specific_dc_field(metadata, "language", sub_elements=["iso"])

def extract_link_site(_, metadata):
    return get_specific_dc_field(metadata, "identifier", sub_elements=["uri"])

def extract_subjects(_, metadata):
    subjects = metadata.xpath(
        "./xoai:element[@name='dc']/xoai:element[@name='subject']/*",
        namespaces=NAMESPACES
    )
    all_subjects = []
    for subj in subjects:
        values = subj.xpath(".//xoai:field[@name='value']/text()", namespaces=NAMESPACES)
        all_subjects.extend([v.strip() for v in values if v.strip()])
    return format_list_to_string(list(set(all_subjects)))

def extract_abstract(_, metadata):
    abstract_paths = [
        "./xoai:element[@name='dc']/xoai:element[@name='description']/xoai:element[@name='abstract']"
        "/xoai:element[@name='pt_BR']/xoai:field[@name='value']/text()",
        "./xoai:element[@name='dc']/xoai:element[@name='description']/xoai:element[@name='abstract']"
        "/xoai:element[@name='none']/xoai:field[@name='value']/text()",
        "./xoai:element[@name='dc']/xoai:element[@name='description']/xoai:element[@name='abstract']"
        "/xoai:element[@name='en']/xoai:field[@name='value']/text()",
        "./xoai:element[@name='dc']/xoai:element[@name='description']"
        "/xoai:element[@name='abstract']/xoai:field[@name='value']/text()"
    ]
    for path in abstract_paths:
        result = metadata.xpath(path, namespaces=NAMESPACES)
        if result:
            text = [r.strip() for r in result if r.strip()]
            if text:
                return text[0]
    return None

def extract_generic_field(_, metadata, field_name):  # para type, title, publisher, description
    return get_specific_dc_field(metadata, field_name)

def extract_date_field(_, metadata, field_name):  # para issued, available, accessioned
    return get_date_field(metadata, field_name)

# Mapeamento dos campos para as funções de extração
FIELD_EXTRACTORS = {
    'identifier_header': extract_identifier,
    'datestamp_header': extract_datestamp,
    'setSpec': extract_set_spec,
    'authors': extract_authors,
    'advisors': extract_advisors,
    'co_advisors': extract_co_advisors,
    'link_doc': extract_link_doc,
    'language': extract_language,
    'link_site': extract_link_site,
    'subjects': extract_subjects,
    'abstract': extract_abstract,
    'title': lambda h, m: extract_generic_field(h, m, "title"),
    'type': lambda h, m: extract_generic_field(h, m, "type"),
    'publisher': lambda h, m: extract_generic_field(h, m, "publisher"),
    'description': lambda h, m: extract_generic_field(h, m, "description"),
    'issued_date': lambda h, m: extract_date_field(h, m, "issued"),
    'available_date': lambda h, m: extract_date_field(h, m, "available"),
    'accessioned_date': lambda h, m: extract_date_field(h, m, "accessioned"),
}


def extract_data_from_xml_file(xml_file_path : str,
                               desired_fields : list[str] = [],
                               logger : logging.Logger=None):
    """
    Extrai dados estruturados de um arquivo XML, retornando uma lista de dicionários.

    Parâmetros:
        - xml_file_path: caminho do arquivo XML
        - desired_fields: lista de nomes dos campos desejados (colunas do dicionário)
        - logger: instância de logger (opcional). Se None, não loga nada.
    """
    if logger:
        logger.info(f"Iniciando extração de dados do arquivo XML: {xml_file_path}")

    try:
        tree = etree.parse(xml_file_path)
        root = tree.getroot()
        if logger:
            logger.info("Arquivo XML analisado com sucesso.")
    except etree.XMLSyntaxError as e:
        if logger:
            logger.error(f"Erro de sintaxe XML no arquivo {xml_file_path}: {e}")
        return []
    except IOError as e:
        if logger:
            logger.error(f"Erro ao ler o arquivo XML {xml_file_path}: {e}")
        return []

    records_data = []
    total_records = 0
    skipped_records = 0

    records = root.xpath('//oai:record', namespaces=NAMESPACES)
    if logger:
        logger.info(f"Número de registros encontrados: {len(records)}")

    for idx, record in enumerate(records, start=1):
        total_records += 1
        header = record.xpath('./oai:header', namespaces=NAMESPACES)
        if not header:
            if logger:
                logger.warning(f"Registro {idx} sem cabeçalho. Ignorado.")
            skipped_records += 1
            continue

        if header[0].get('status') == 'deleted':
            if logger:
                logger.info(f"Registro {idx} marcado como deletado. Ignorado.")
            skipped_records += 1
            continue

        header = header[0]
        metadata_xoai = record.xpath('./oai:metadata/xoai:metadata', namespaces=NAMESPACES)
        metadata_xoai = metadata_xoai[0] if metadata_xoai else None

        record_info = {}
        record_info["source_xml_file"] = os.path.basename(xml_file_path)

        for field, extractor in FIELD_EXTRACTORS.items():
            if (not desired_fields) or (field in desired_fields):
                try:
                    record_info[field] = extractor(header, metadata_xoai)
                except Exception as e:
                    if logger:
                        logger.warning(f"Erro ao extrair campo '{field}' do registro {idx}: {e}")
                    record_info[field] = None

        records_data.append(record_info)

    if logger:
        logger.info(
            f"Extração concluída. Total: {total_records}, "
            f"Processados: {total_records - skipped_records}, "
            f"Ignorados: {skipped_records}"
        )

    return records_data


def get_col_to_name_csv_file():
    if os.path.exists(COL_TO_NAME_CSV_FILE_PATH):
        with open(COL_TO_NAME_CSV_FILE_PATH, "r", encoding="utf-8") as f:
            reader = csv.reader(f)
            next(reader)  # pular cabeçalho
            return {linha[0]: linha[1] for linha in reader}
    else:
        return {}
    
dic_col_to_name = get_col_to_name_csv_file()

def save_col_to_name_csv_file(dic_col_to_name : dict[str,str]):
    with open(COL_TO_NAME_CSV_FILE_PATH, "w", newline='', encoding='utf-8') as f:
        writer = csv.writer(f)
        writer.writerow(["col", "location"])  # cabeçalho (opcional)
        for col, location in dic_col_to_name.items():
            writer.writerow([col, location])

def extract_year(date_str):
    if pd.isnull(date_str):
        return None
    match = re.match(r'(\d{4})', str(date_str))
    if match:
        return match.group(1)
    return ''

def format_type(work_type : str) -> str:
    dic_types = {
        '':'Não Especificado',
        '(Dissertação (mestrado)':'Dissertação Mestrado',
        '(HTML & PDF & DOC)':'HTML-PDF-DOC',
        'Article':'Artigo',
        'D':'D',
        'Disertação (Mestrado)':'Dissertação Mestrado',
        'Disseratação (Mestrado)':'Dissertação Mestrado',
        'Dissertacão (Mestrado)':'Dissertação Mestrado',
        'Dissertaão (Mestrado)':'Dissertação Mestrado',
        'Dissertaçao (Mestrado)':'Dissertação Mestrado',
        'Dissertaçaõ (Mestrado)':'Dissertação Mestrado',
        'Dissertaçào (Mestrado)':'Dissertação Mestrado',
        'Dissertação':'Dissertação',
        'Dissertação  (Mestraddo)':'Dissertação Mestrado',
        'Dissertação  (Mestrado)':'Dissertação Mestrado',
        'Dissertação ( mestrado )':'Dissertação Mestrado',
        'Dissertação ( mestrado)':'Dissertação Mestrado',
        'Dissertação (Administração)':'Dissertação',
        'Dissertação (Dissertação)':'Dissertação',
        'Dissertação (Doutorado)':'Dissertação Doutorado',
        'Dissertação (Mesrtado)':'Dissertação Mestrado',
        'Dissertação (Mesrtrado)':'Dissertação Mestrado',
        'Disertação (Mestrado)':'Dissertação Mestrado',
        'Dissertação (Mestado)':'Dissertação Mestrado',
        'Dissertação (Mestraddo)':'Dissertação Mestrado',
        'Dissertação (Mestrado )':'Dissertação Mestrado',
        'Dissertação (Mestrado acadêmico)':'Dissertação Mestrado',
        'Dissertação (Mestrado profissional)':'Dissertação Mestrado Profissional',
        'Dissertação (Mestrado)':'Dissertação Mestrado',
        'Dissertação (Mestrrado)':'Dissertação Mestrado',
        'Dissertação (Metrado)':'Dissertação Mestrado',
        'Dissertação (mestrado)':'Dissertação Mestrado',
        'Dissertação [mesterado)':'Dissertação Mestrado',
        'Dissertação [mestrado)':'Dissertação Mestrado',
        'Dissertação mestrado)':'Dissertação Mestrado',
        'Dissertação {mestrado)':'Dissertação Mestrado',
        'Dissertação( mestrado)':'Dissertação Mestrado',
        'Dissertação(Mestrado)':'Dissertação Mestrado',
        'Dissertação(Mestrdo)':'Dissertação Mestrado',
        'Dissertaçãom (Mestrado)':'Dissertação Mestrado',
        'Dissertaçção (Mestrado)':'Dissertação Mestrado',
        'Dissetação (Mestrado)':'Dissertação Mestrado',
        'Disssertação (Mestrado)':'Dissertação Mestrado',
        'Editorial design and revision by Beatriz Stephanie Ribeiro':'Editorial design and revision by Beatriz Stephanie Ribeiro', 
        'Monografia (Especialização em Planejamento e Gestão em Defesa Civil)':'Monografia',
        'Monografia de especialização':'Monografia',
        'Other: Editorial design and revision by Beatriz Stephanie Ribeiro':'Outros',
        'Relatorio (Pós-doutorado)':'Relatorio Pós-Doutorado',
        'Relatório (Pós-Doutorado)':'Relatorio Pós-Doutorado',
        'Relatório Técnico (Mestrado profissional)':'Relatório Técnico Mestrado Profissional',
        'Relatório de Estágio Extr':'Relatório Estágio Extr',
        'TCC (graduação Arquitetura e Urbanismo)':'TCC',
        'TCC (graduação em Agronomia )':'TCC',
        'TCC (graduação em Agronomia)':'TCC',
        'TCC (graduação em Arquitetura e Urbanismo )':'TCC',
        'TCC (graduação em Arquitetura e Urbanismo)':'TCC',
        'TCC (graduação em Arquitetura e Urbansimo)':'TCC',
        'TCC (graduação em Biblioteconomia)':'TCC',
        'TCC (graduação em Emfermagem)':'TCC',
        'TCC (graduação em Enfermagem )':'TCC',
        'TCC (graduação em Enfermagem)':'TCC',
        'TCC (graduação em Engenharia de Aquicultura )':'TCC',
        'TCC (graduação em Engenharia de Aquicultura)':'TCC',
        'TCC (graduação em Engenharia de Aqüicultura)':'TCC',
        'TCC (graduação em Serviço Social)':'TCC',
        'TCC (graduação)':'TCC',
        'TCC (graduaçãoem Agronomia)':'TCC',
        'TCC1 (graduação)':'TCC',
        'TCC2 (Graduação em Arquitetura e Urbanismo)':'TCC',
        'TCCP (especialização)':'TCCP Especialização',
        'TCCes':'TCC',
        'TCCgrad':'TCC',
        'TCCresid':'TCC',
        'TESE (Doutor)':'Tese Doutorado',
        'TESE (Doutorado)':'Tese Doutorado',
        'Tese  (Doutorado)':'Tese Doutorado',
        'Tese (Dissertação)':'Tese;Dissertação',
        'Tese (Dotutorado)':'Tese Doutorado',
        'Tese (Dourado)':'Tese Doutorado',
        'Tese (Dourorado)':'Tese Doutorado',
        'Tese (Doutarado)':'Tese Doutorado',
        'Tese (Doutorado profissional)':'Tese Doutorado Profissional',
        'Tese (Doutorado)':'Tese Doutorado',
        'Tese (Doutordo)':'Tese Doutorado',
        'Tese (Doutotado)':'Tese Doutorado',
        'Tese (Doutrado)':'Tese Doutorado',
        'Tese (Dutorado)':'Tese Doutorado',
        'Tese (Livre Docência)':'Tese Livre Docência',
        'Tese (Livre docencia)':'Tese Livre Docência',
        'Tese (Livre-docencia)':'Tese Livre Docência',
        'Tese (Mestrado)':'Tese Mestrado',
        'Tese (doutorado)':'Tese Doutorado',
        'Tese - (Doutorado)':'Tese Doutorado',
        'Tese Doutorado)':'Tese Doutorado',
        'Tese [doutorado)':'Tese Doutorado',
        'Tese elaborada em regime de cotutela entre o Programa de Pós Graduação em Engenharia de Alimentos da Universidade Federal de Santa Catarina e a Escola de Doutorado de Ciências da Engenharia (SPI Oniris)':'Tese',
        'Tese {doutorado)':'Tese Doutorado',
        'Tese(Doutorado)':'Tese Doutorado',
        'Trabalho de Conclusao de Curso':'TCC',
        'Trabalho de Conclusão (Graduação)':'TCC',
        'Trabalho de Conclusão de Curso':'TCC',
        'Trabalho de Conclusão de Curso  (graduação em Engenharia de Aquicultura )':'TCC',
        'article':'Artigo',
        'dissertacao':'Dissertação',
        'e-Book':'e-book',
        'eBook':'e-book',
        'filme':'Filme', 
        'image':'Imagem',
        'imagem':'Imagem',
        'other':'Outros',
        'relatorio':'Relatório',
        'report':'Relatório',
        'tese':'Tese',
        'tese (Doutorado)':'Tese Doutorado',
    }
    if work_type in dic_types.keys():
        return dic_types[work_type].upper().strip()
    return work_type.upper().strip()

def get_text_trail(url : str,
                   logger : logging.Logger|None = None,
                   timeout : int|float = 65) -> str:
    try:
        response = requests.get(url,timeout=timeout)
        response.raise_for_status()
    except requests.exceptions.Timeout:
        if logger:
            logger.error("Erro: A requisição excedeu o tempo limite de 65 segundos.",exc_info=True)
        return ''
    except requests.exceptions.HTTPError as errh:
        if logger:
            logger.error(f"Erro HTTP: {errh}",exc_info=True)
        return ''
    except requests.exceptions.RequestException as err:
        if logger:
            logger.error(f"Erro na requisição: {err}",exc_info=True)
        return ''
    except Exception as e:
        if logger:
            logger.error(f'Erro desconhecido ocorreu na hora de processar requisição --> {e.__class__.__name__}: {str(e)}',exc_info=True)
        return ''
    else:
        soup = BeautifulSoup(response.text, "html.parser")
        try:
            header_div = soup.find("div", id="ds-header")
            if not header_div:
                if logger:
                    logger.warning(f'Não foi possível encontrar header no HTML')
                return ''

            trail_items = header_div.find_all("li")
            if len(trail_items) < 3:
                if logger:
                    logger.warning(f'Não foi possível encontrar a trilha de localização no HTML')
                return ''

            # We expect the full path to be something like this:
            # "Teses e Dissertações -> Programa de Pós-Graduação em Engenharia de Automação e Sistemas"
            # Or even:
            # "Acervos -> Campus Florianópolis -> PROPESQ (Pró-Reitoria de Pesquisa) -> Programa de Iniciação Científica e Tecnológica da UFSC -> Seminário de Iniciação Científica e Tecnológica da UFSC -> 2023 -> Iniciação Científica - PIBIC e Programa Voluntário -> Ciências Exatas, da Terra e Engenharias -> Departamento de Automação e Sistemas"
            str_full_location = ' -> '.join([trail_item.get_text(strip=True) for trail_item in trail_items if trail_item.get_text(strip=True) not in ['Repositório Institucional da UFSC',
                                                                                                                                                      'DSpace Home',
                                                                                                                                                      'Ver item',
                                                                                                                                                      'View Item']])
            return str_full_location

        except Exception as e:
            if logger:
                logger.error(f'Erro desconhecido na hora de processar resposta da requisição --> {e.__class__.__name__}: {str(e)}',exc_info=True)
            return ''
        
def get_origin_info_from_col(col : str,
                             logger : logging.Logger|None = None) -> dict[str]:
    
    # Esperamos que col seja algo como "col_123456789_75030" ...

    col_numbers = col.split('_')
    col_numbers.remove('col')
    first_number = col_numbers[0]
    second_number = col_numbers[1]

    link = f'https://repositorio.ufsc.br/handle/{first_number}/{second_number}'

    str_full_location = get_text_trail(link,logger)

    if logger:
        logger.info(f'Link para coletar info {link}')
        if str_full_location:
            logger.info(f'Localização completa encontrada para {col}: {str_full_location}')

    return str_full_location

def get_academic_work_id_locator(dic_xml_record : dict) -> list[str]:
    if isinstance(dic_xml_record,dict):
        header = dic_xml_record.get('header')
        if header:
            setSpec = header.get('setSpec')
            if setSpec:
                if isinstance(setSpec,list):
                    return [item.strip() for item in setSpec if re.search(r'col\_\d+\_\d+',item)]
                elif isinstance(setSpec,str):
                    return [setSpec.strip()] if re.search(r'col\_\d+\_\d+',setSpec) else []
                else:
                    return []
        return []
    else:
        raise TypeError(f'Input type is not dict, as expected. Current input type is {type(dic_xml_record).__name__}.')

def get_academic_work_location(id_loc : str, logger : logging.Logger|None = None) -> str:
    global dic_col_to_name
    if id_loc not in dic_col_to_name.keys():
        if logger:
            logger.info(f'Coletando localização por nome do id "{id_loc}"')
        full_location = get_origin_info_from_col(id_loc,logger)
        if full_location:
            dic_col_to_name[id_loc] = full_location
            save_col_to_name_csv_file(dic_col_to_name)
            return full_location        
    else:
        # if logger:
        #     logger.info(f'Coletando informação de localização pelo dicionário armazenado com id "{id_loc}"')
        full_location = dic_col_to_name[id_loc]
        return full_location
    if logger:
        logger.warning(f'Não foi possível identificar localização completa para id {str(id_loc)}')
    return ''

def get_academic_work_first_community(full_location : str) -> str:
    return full_location.split('->')[0].strip()

def get_academic_work_last_collection(full_location : str) -> str:
    return full_location.split('->')[-1].strip()

def insert_location_into_df(df : pd.DataFrame,logger : logging.Logger|None = None) -> pd.DataFrame:
    setSpecs = df['setSpec'].to_list()
    df['full_locations'] = [get_academic_work_location(value,logger=logger) if value else '' for value in setSpecs]    
    df['first_com'] = [get_academic_work_first_community(value) if value else '' for value in df['full_locations']]
    df['last_col'] = [get_academic_work_last_collection(value) if value else '' for value in df['full_locations']]

    return df

def split_full_location(full_location : str) -> list[str]:
    splitted_full_location = [loc.strip() for loc in full_location.split('->') if len(loc.strip())>3]
    return splitted_full_location

def treat_locations(full_location : str, reverse : bool = True) -> list[str]:
    full_location = re.sub(r'Curso de|Programa de Pós-Graduação em','',full_location,flags=re.IGNORECASE).strip()
    location_elements = [format_text(loc,special_treatment=True).strip() for loc in split_full_location(full_location)]    
    if reverse:
        location_elements.reverse() # Trabalharemos com prioridade para as últimas localizações (ler de trás para frente)
    return location_elements

def get_curso_from_full_location(full_location : str,courses : list[str]|None = None) -> str:
    # Só coletamos correspondências totais (==) em algum elemento entre os "->" com os cursos catalogados pela UFSC
    # Teses e Dissertações -> Programa de Pós-Graduação em Oceanografia se torna uma lista com ["Teses e Dissertações", "Programa de Pós-Graduação em Oceanografia"]
    # Depois o último elemento da lista tem "Programa de Pós-Graduação em" removido, ficando só com o nome do curso exatamente.
    # Na hora da comparação ambos os elementos são formatados.
    location_elements = treat_locations(full_location) # Trabalharemos com prioridade para as últimas localizações (ler de trás para frente)    
    for element in location_elements:            
        for curso_ufsc in [c for c in courses if len(c) <= len(element)]:
            curso_ufsc_formatted = format_text(curso_ufsc,special_treatment=True)
            if curso_ufsc_formatted == element:
                return curso_ufsc
    return ''

def get_list_of_curso_from_full_location(list_of_courses : list[str],
                                         list_of_full_locations : list[str],
                                         ufsc_courses) -> list[str]:
    courses = []
    for course,full_location in zip(list_of_courses,list_of_full_locations):
        if course.strip() == '':
            courses.append(get_curso_from_full_location(full_location=full_location,courses=ufsc_courses))
        else:
            courses.append(course)
    return courses

def split_description(description : str) -> list[str]:
    splitted_description = [desc.strip() for desc in re.split(r'\s\-\s|\,|\.',description) if len(desc.strip())>3]
    return splitted_description

def treat_descriptions(description : str, reverse : bool = True) -> list[str]:
    description = re.sub(r'Curso de|Programa de Pós-Graduação em','',description,flags=re.IGNORECASE).strip()   
    description_elements = [format_text(element,special_treatment=True).strip() for element in split_description(description)]
    if reverse:
        description_elements.reverse()
    return description_elements

def get_curso_from_description(description : str, courses : list[str]|None = None) -> str:
    description_elements = treat_descriptions(description)
    for element in description_elements:
        for curso_ufsc in [c for c in courses if len(c) <= len(element)]:
            curso_ufsc_formatted = format_text(curso_ufsc,special_treatment=True)
            if curso_ufsc_formatted == element:
                return curso_ufsc
    return ''

def insert_curso_into_df(df : pd.DataFrame) -> pd.DataFrame:
    try:
        ufsc_courses = CursosUFSC().get_cursos()
    except Exception as e:
        ufsc_courses = None
    df['course'] = [get_curso_from_description(value,ufsc_courses) if value else '' for value in df['description'].to_list()]
    df['course'] = get_list_of_curso_from_full_location(list_of_courses=df['course'].to_list(),
                                                        list_of_full_locations=df['full_locations'].to_list(),
                                                        ufsc_courses=ufsc_courses)
    df['course'] = df['course'].str.upper()
    return df


def force_type_curso_from_description(list_type_cursos : list[str],
                                      list_descriptions : list[str]) -> list[str]:
    forced_list_type_courses = []
    for type_course,description in zip(list_type_cursos,list_descriptions):
        description = description.strip()
        if type_course == '' and description: # Se não tiver tipo_curso, mas tiver uma descrição
            if re.search(r'^tese|^dissertação',description,re.IGNORECASE):
                forced_list_type_courses.append('POS')
            elif re.search(r'^tcc|^trabalho_conclusao_curso|^trabalho_de_conclusao_de_curso|^trabalho_conclusao_de_curso|^pfc|^projeto_fim_de_curso|^projeto_de_fim_de_curso',description,re.IGNORECASE):
                forced_list_type_courses.append('GRAD')
            else:
                forced_list_type_courses.append('')
        else:
            forced_list_type_courses.append(type_course)
    return forced_list_type_courses

def insert_type_curso_based_on_description_into_df(df : pd.DataFrame) -> pd.DataFrame:
    if 'type_course' not in df.keys():
        df['type_course'] = ''    
    # Tentar coletar tipo de curso por full_location e por curso não parece ser boa opção
    # Complexidade no full_loc não garante precisão e existem cursos que tem tanto na GRAD quanto na POS
    # df['type_course'] = force_type_curso_from_full_location(list_type_cursos=list_type_courses,list_full_locations=df['full_locations'].to_list())
    # df['type_course'] = force_type_curso_from_courses(list_type_cursos=df['type_course'].to_list(),list_courses=df['course'].to_list())
    df['type_course'] = force_type_curso_from_description(list_type_cursos=df['type_course'].to_list(),
                                                          list_descriptions=df['description'].to_list())
    return df

def insert_type_course_based_on_type_into_df(df: pd.DataFrame, logger: logging.Logger = None) -> pd.DataFrame:
    """
    Adiciona coluna 'type_course' com valores 'POS', 'GRAD' ou string vazia ("")
    com base no conteúdo da coluna 'type'.
    """

    # Inicializa com string vazia
    df['type_course'] = ""

    # Expressões regulares corrigidas com grupos de NÃO captura
    regex_pos = re.compile(
        r'^tese(?:s)?\b|^dissertacao\b|^dissertacoes\b|_mestrado\b|_doutorado\b|mestrado|doutorado|tese|dissertacao|dissertação',
        flags=re.IGNORECASE)

    regex_grad = re.compile(
        r'^tcc\b|^tcc\_|^tccp\b|^tccp\_',
        flags=re.IGNORECASE)

    # Aplica POS
    df.loc[
        df['type'].fillna('').str.lower().str.contains(regex_pos),
        'type_course'
    ] = 'POS'

    # Aplica GRAD
    df.loc[
        df['type'].fillna('').str.lower().str.contains(regex_grad),
        'type_course'
    ] = 'GRAD'

    if logger:
        logger.info(f"Coluna 'type_course' adicionada com valores únicos: {df['type_course'].unique().tolist()}")

    return df


def get_campus_from_description(list_campus : list[str],
                                list_descriptions : list[str]) -> list[str]:
    forced_list_campus = []
    
    dic_centros_campus = {'FLN':[format_text(centro,special_treatment=True) for centro in DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS['FLN'].keys()],
                          'JOI':[format_text(centro,special_treatment=True) for centro in DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS['JOI'].keys()],
                          'CUR':[format_text(centro,special_treatment=True) for centro in DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS['CUR'].keys()],
                          'BNU':[format_text(centro,special_treatment=True) for centro in DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS['BNU'].keys()],
                          'ARA':[format_text(centro,special_treatment=True) for centro in DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS['ARA'].keys()]}

    for campus,description in zip(list_campus,list_descriptions):
        if campus == '' and description.strip():
            formatted_description = format_text(description,special_treatment=True)
            if 'campus_florianopolis' in formatted_description:
                forced_list_campus.append('FLN')
            elif 'campus_ararangua' in formatted_description:
                forced_list_campus.append('ARA')
            elif 'campus_blumenau' in formatted_description:
                forced_list_campus.append('BNU')
            elif 'campus_curitibanos' in formatted_description:
                forced_list_campus.append('CUR')
            elif 'campus_joinville' in formatted_description:
                forced_list_campus.append('JOI')
            elif re.search(r'florianopolis\_\d{4}$',formatted_description):
                forced_list_campus.append('FLN')
            else:
                campus_status = False
                for campus_dic in dic_centros_campus.keys():
                    for formatted_centro in dic_centros_campus[campus_dic]:
                        if formatted_centro in formatted_description:
                            campus_status = True
                            forced_list_campus.append(campus_dic)
                            break
                    if campus_status:
                        break
                if not campus_status:
                    forced_list_campus.append('')
        else:
            forced_list_campus.append(campus)
    return forced_list_campus

def insert_campus_from_description_into_df(df : pd.DataFrame) -> pd.DataFrame:
    # Não podemos usar insert_campus_from_cursos_ufsc() porque há cursos com o mesmo tipo que são oferecidos em diferentes campus
    # Como Agronomia (GRAD) em FLN e CUR / Eng. Controle e Automação (GRAD) FLN e BNU / Materiais (GRAD) / Medicina (GRAD), etc
    # df = insert_campus_from_cursos_ufsc(df)
    df['campus'] = ''
    df['campus'] = get_campus_from_description(list_campus=df['campus'].to_list(),list_descriptions=df['description'].to_list())

    return df


def get_list_of_centro_from_description(list_centros : list[str],
                                list_descriptions : list[str]) -> list[str]:
    forced_list_centros = []
    dic_formatted_centros = {format_text('Centro Tecnológico, de Ciências exatas e Educação',special_treatment=True).strip():'CTE',
                             format_text('Centro de Ciências Físicas e Matemáticas',special_treatment=True).strip():'CFM',
                             format_text('Centro de Ciências, Tecnologias e Saúde',special_treatment=True).strip():'CTS',
                             format_text('Centro de Filosofia e Ciências Humanas',special_treatment=True).strip():'CFH',
                             format_text('Centro de Comunicação e Expressão',special_treatment=True).strip():'CCE',
                             format_text('Centro Tecnológico de Joinville',special_treatment=True).strip():'CTJ',
                             format_text('Centro de Ciências da Educação',special_treatment=True).strip():'CED',
                             format_text('Centro de Ciências Biológicas',special_treatment=True).strip():'CCB',
                             format_text('Centro de Ciências Jurídicas',special_treatment=True).strip():'CCJ',
                             format_text('Centro de Ciências Agrárias',special_treatment=True).strip():'CCA',
                             format_text('Centro de Ciências da Saúde',special_treatment=True).strip():'CCS',
                             format_text('Centro de Ciências Rurais',special_treatment=True).strip():'CCR',
                             format_text('Centro Socio-econômico',special_treatment=True).strip():'CSE', # Primeira variação do CSE
                             format_text('Centro Socioeconômico',special_treatment=True).strip():'CSE', # Segunda variação do CSE
                             format_text('Centro de Desportos',special_treatment=True).strip():'CDS',                             
                             format_text('Centro Tecnológico',special_treatment=True).strip():'CTC'}
    for centro,description in zip(list_centros,list_descriptions):
        if centro == '' and description.strip():
            formatted_description = format_text(description,special_treatment=True).replace('centro_de_','centro_').strip()
            centro_status = False
            for formatted_centro in dic_formatted_centros.keys():
                if formatted_centro.replace('centro_de_','centro_') in formatted_description:
                    forced_list_centros.append(dic_formatted_centros[formatted_centro])
                    centro_status = True
                    break
            if not centro_status:
                forced_list_centros.append('')
        else:
            forced_list_centros.append(centro)
    return forced_list_centros

def get_list_of_centro_from_campus(list_centros : list[str],
                            list_campus : list[str]) -> list[str]:
    forced_centros = []
    campus_with_just_one_centro = [campi for campi in DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS if len(DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS[campi].keys()) == 1]
    for centro, campi in zip(list_centros,list_campus):
        if centro == '' and campi in campus_with_just_one_centro:
            centro_full_name = list(DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS[campi].keys())[0]
            forced_centros.append(DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS[campi][centro_full_name])
        else:
            forced_centros.append(centro)
    return forced_centros

def insert_centro_into_df(df : pd.DataFrame) -> pd.DataFrame:
    # Não podemos usar insert_centro_from_cursos_ufsc() porque há cursos com o mesmo tipo que são oferecidos em diferentes centros
    # Como Eng. Controle e Automação no CTC (FLN) e CTE (BNU)
    # df = insert_centro_from_cursos_ufsc(df)
    df['centro'] = ''
    df['centro'] = get_list_of_centro_from_campus(list_centros=df['centro'].to_list(),
                                           list_campus=df['campus'].to_list())
    df['centro'] = get_list_of_centro_from_description(list_centros=df['centro'].to_list(),
                                               list_descriptions=df['description'].to_list())
    return df


def get_centro_from_campus_and_course(list_centros : list[str],
                                      list_campus : list[str],
                                      list_courses : list[str]) -> list[str]:    
    
    forced_list_centros = []
    for centro,campus,curso in zip(list_centros,list_campus,list_courses):
        if centro == '':
            if campus.strip() and curso.strip():
                try:
                    desired_centro = DIC_CAMPUS_CURSOS_CENTROS_SIGLAS[campus][curso]
                    forced_list_centros.append(desired_centro)
                except Exception as e:
                    forced_list_centros.append('')
            else:
                forced_list_centros.append('')
        else:
            forced_list_centros.append('')
    return forced_list_centros

def insert_centro_from_campus_and_course_into_df(df : pd.DataFrame) -> pd.DataFrame:
    df['centro'] = get_centro_from_campus_and_course(list_centros=df['centro'].to_list(),
                                                     list_campus=df['campus'].to_list(),
                                                     list_courses=df['course'].to_list())

def get_gender_by_name(name: str) -> str:
    if name:
        name = name.split()[0]
        genero = br_gender_info.get_gender(name)
        if genero == 'Male':
            return 'M'
        elif genero == 'Female':
            return 'F'
    return ''

def get_author_first_name(author_name : str) -> str:
    if ',' in author_name:
        comma_index = author_name.index(',')
        if comma_index + 1 < len(author_name):
            author_name = author_name[comma_index+1:].strip()
            if author_name:
                return author_name.split()[0]
    return ''

def get_authors_first_names(authors : str) -> list[str]:
    authors_first_names = []
    splitted_authors = authors.split(';')
    for author in splitted_authors:
        author_first_name = get_author_first_name(author)
        if author_first_name:
            authors_first_names.append(author_first_name)
    return authors_first_names

def get_authors_gender_by_name(authors : str) -> str:
    authors_first_names = get_authors_first_names(authors)
    if authors_first_names:
        genders = ''
        for author_first_name in authors_first_names:
            author_gender_name = get_gender_by_name(author_first_name)
            if author_gender_name not in genders:
                if genders:
                    genders += f',{author_gender_name}'
                else:
                    genders = author_gender_name
        if 'F' in genders and 'M' in genders: # Deixar valor padrão quando tiver os 2 gêneros
            genders = 'F,M'
        return genders
    return ''

def get_gender_by_name_list_for_df(list_authors_names : list[str]) -> list[str]:
    return [get_authors_gender_by_name(authors_name) for authors_name in list_authors_names]

def insert_gender_by_name_into_df(df : pd.DataFrame) -> pd.DataFrame:
    df['gender_name'] = get_gender_by_name_list_for_df(list_authors_names=df['authors'].to_list())
    return df

def get_list_of_campus_from_centro(list_of_centros : list[str],
                                   list_of_campus : list[str]) -> list[str]:
    forced_campus = []
    for campi, centro in zip(list_of_campus,list_of_centros):
        if campi == '' and centro.strip():
            status_campi = False
            for desired_campi in DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS.keys():
                if centro in [DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS[desired_campi][centro_full_name] for centro_full_name in DIC_CAMPUS_CENTROS_COMPLETO_E_SIGLAS[desired_campi].keys()]:
                    forced_campus.append(desired_campi)
                    status_campi = True
                    break
            if not status_campi:
                forced_campus.append('')
        else:
            forced_campus.append(campi)
    return forced_campus

def adjust_campus_from_centro(df : pd.DataFrame) -> pd.DataFrame:    
    df['campus'] = get_list_of_campus_from_centro(list_of_centros=df['centro'].to_list(),
                                                  list_of_campus=df['campus'].to_list())
    return df

def insert_new_columns_into_df(df : pd.DataFrame,
                               logger:logging.Logger|None=None) -> pd.DataFrame:
    try:
        txt_log_error = 'Problema na inserção de colunas no dataframe com função fillna()'
        df.fillna('',inplace=True)
        txt_log_error = 'Problema na inserção de colunas no dataframe com função astype()'
        df = df.astype(str)

        txt_log_error = 'Problema na inserção de colunas no dataframe com formatação da coluna type com função format_type()'
        df['type'] = df['type'].apply(format_type)

        txt_log_error = 'Problema na inserção de colunas no dataframe com função insert_gender_by_name_into_df()'
        df = insert_gender_by_name_into_df(df)
        
        txt_log_error = 'Problema na inserção de colunas no dataframe com função insert_location_into_df()'
        df = insert_location_into_df(df,logger=logger)

        txt_log_error = 'Problema na inserção de colunas no dataframe com função insert_curso_into_df()'
        df = insert_curso_into_df(df)
        
        txt_log_error = 'Problema na inserção de colunas no dataframe com função insert_type_course_based_on_type_into_df()'
        df = insert_type_course_based_on_type_into_df(df,logger=logger)
        
        txt_log_error = 'Problema na inserção de colunas no dataframe com função insert_type_curso_based_on_description_into_df()'
        df = insert_type_curso_based_on_description_into_df(df)
        
        txt_log_error = 'Problema na inserção de colunas no dataframe com função insert_campus_from_description_into_df()'
        df = insert_campus_from_description_into_df(df)
        
        txt_log_error = 'Problema na inserção de colunas no dataframe com função insert_centro_from_description_into_df()'
        df = insert_centro_into_df(df)

        txt_log_error = 'Problema na inserção de colunas no dataframe com função adjust_campus_from_centro()'
        df = adjust_campus_from_centro(df)
        
        txt_log_error = 'Problema na inserção de colunas no dataframe com construção da coluna year com função extract_year()'
        df['year'] = df['issued_date'].apply(extract_year)
    except Exception as e:
        if logger:
            logger.error(f'{txt_log_error} --> "{e}"',exc_info=True)
    return df

def order_df_columns(df : pd.DataFrame) -> pd.DataFrame:
    # Ordenando colunas conforme desejado
    column_order = [
        'identifier_header', 'datestamp_header', 'setSpec', 'title', 'authors', 'advisors', 'co_advisors',
        'issued_date', 'available_date', 'accessioned_date',
        'language', 'subjects', 'type', 'publisher', 'description',
        'abstract', 'link_site', 'link_doc', 'source_xml_file'
    ]
    existing_ordered_cols = [col for col in column_order if col in df.columns]
    remaining_cols = [col for col in df.columns if col not in existing_ordered_cols]
    final_columns = existing_ordered_cols + remaining_cols
    df = df[final_columns]
    return df

def transform_df(df : pd.DataFrame,logger:logging.Logger|None=None) -> pd.DataFrame:
    df = insert_new_columns_into_df(df,logger)
    if logger:
        logger.info('Tentativa de inserção de colunas finalizada')

    df = order_df_columns(df)    
    if logger:
        logger.info('Tentativa de ordenação de colunas finalizada')

    return df
